"""
Test torch native ops.
"""

import _setup_test_env  # noqa
import os
from unittest import SkipTest
from typing import Optional, Union, Sequence, List
import numpy
from numpy.testing import assert_almost_equal, assert_allclose
import scipy.ndimage
import torch
from returnn.torch.util.native_op import (
    make_fast_baum_welch_op,
    fast_baum_welch,
    fast_viterbi,
    get_ctc_fsa_fast_bw,
    edit_distance,
    optimal_completion_edit_distance,
    optimal_completion_edit_distance_per_successor,
    optimal_completion_edit_distance_per_successor_via_next_edit_distance,
    edit_distance_via_next_edit_distance_row,
)
from returnn.util.fsa import FastBwFsaShared, get_ctc_fsa_fast_bw as get_ctc_fsa_fast_bw_np
from fsa_utils import py_baum_welch, py_viterbi
from numpy_ref_edit_distance import edit_distance_ref_np, edit_distance_ref_np_b1


os.environ["RETURNN_NATIVE_CODE_COMPILER_VERBOSE"] = "1"


def test_compile():
    # In case CUDA is available, it will use nvcc to compile, and compiles the op with both CPU and CUDA support.
    make_fast_baum_welch_op(compiler_opts=dict(verbose=True))


def test_compile_cpu_only():
    # This forces CPU-only compilation, using c++, even if CUDA is available.
    make_fast_baum_welch_op(compiler_opts=dict(verbose=True), with_cuda=False)


def test_compile_with_cuda():
    if not torch.cuda.is_available():
        raise SkipTest("CUDA not available")
    # Explicitly compile with CUDA support.
    make_fast_baum_welch_op(compiler_opts=dict(verbose=True), with_cuda=True)


def test_FastBaumWelch():
    print("Make op...")
    op = make_fast_baum_welch_op(
        compiler_opts=dict(verbose=True)
    )  # will be cached, used inside :func:`fast_baum_welch`
    print("Op:", op)
    n_batch = 3
    seq_len = 5
    n_classes = 10

    fsa = FastBwFsaShared()
    fsa.add_inf_loop(state_idx=0, num_emission_labels=n_classes)
    fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
    edges = torch.tensor(fast_bw_fsa.edges, dtype=torch.int32)
    weights = torch.tensor(fast_bw_fsa.weights, dtype=torch.float32)
    start_end_states = torch.tensor(fast_bw_fsa.start_end_states, dtype=torch.int32)
    am_scores_np = numpy.random.RandomState(42).normal(size=(seq_len, n_batch, n_classes)).astype("float32")
    am_scores = torch.tensor(am_scores_np, dtype=torch.float32)  # in -log space
    float_idx_np = numpy.ones((seq_len, n_batch), dtype="float32")
    float_idx = torch.ones((seq_len, n_batch), dtype=torch.float32)
    print("Construct call...")
    fwdbwd, obs_scores = fast_baum_welch(
        am_scores=am_scores, seq_mask=float_idx, edges=edges, weights=weights, start_end_states=start_end_states
    )
    print("Done.")
    print("Eval:")
    fwdbwd_np, score = fwdbwd.numpy(), obs_scores.numpy()
    print("score:", score)
    print("Baum-Welch soft alignment:")
    print(repr(fwdbwd_np))
    fwdbwd_np2, score2 = py_baum_welch(
        am_scores=am_scores_np,
        float_idx=float_idx_np,
        edges=fast_bw_fsa.edges,
        weights=fast_bw_fsa.weights,
        start_end_states=fast_bw_fsa.start_end_states,
    )
    print("ref score:", score2)
    print("ref Baum-Welch soft alignment:")
    print(repr(fwdbwd_np2))
    numpy.testing.assert_allclose(score, score2, rtol=1e-5)
    numpy.testing.assert_allclose(fwdbwd_np, fwdbwd_np2, rtol=1e-5)


def test_fast_bw_uniform():
    print("Make op...")
    op = make_fast_baum_welch_op(
        compiler_opts=dict(verbose=True)
    )  # will be cached, used inside :func:`fast_baum_welch`
    # args: (am_scores, edges, weights, start_end_states, float_idx, state_buffer)
    print("Op:", op)
    n_batch = 3
    seq_len = 7
    n_classes = 5

    fsa = FastBwFsaShared()
    for i in range(n_classes):
        fsa.add_edge(i, i + 1, emission_idx=i)  # fwd
        fsa.add_edge(i + 1, i + 1, emission_idx=i)  # loop
    assert n_classes <= seq_len
    fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
    edges = torch.tensor(fast_bw_fsa.edges, dtype=torch.int32)
    weights = torch.tensor(fast_bw_fsa.weights, dtype=torch.float32)
    start_end_states = torch.tensor(fast_bw_fsa.start_end_states, dtype=torch.int32)
    am_scores = numpy.ones((seq_len, n_batch, n_classes), dtype="float32") * numpy.float32(1.0 / n_classes)
    am_scores = -numpy.log(am_scores)  # in -log space
    am_scores = torch.tensor(am_scores, dtype=torch.float32)
    float_idx = torch.ones((seq_len, n_batch), dtype=torch.float32)
    print("Construct call...")
    fwdbwd, obs_scores = fast_baum_welch(
        am_scores=am_scores, seq_mask=float_idx, edges=edges, weights=weights, start_end_states=start_end_states
    )
    print("Done.")
    print("Eval:")
    fwdbwd, score = fwdbwd.numpy(), obs_scores.numpy()
    print("score:")
    print(repr(score))
    assert score.shape == (seq_len, n_batch)
    bw = numpy.exp(-fwdbwd)
    print("Baum-Welch soft alignment:")
    print(repr(bw))
    assert bw.shape == (seq_len, n_batch, n_classes)

    if seq_len == n_classes:
        print("Extra check identity...")
        for i in range(n_batch):
            assert_almost_equal(numpy.identity(n_classes), bw[:, i])
    if seq_len == 7 and n_classes == 5:
        print("Extra check ref_align (7,5)...")
        assert_allclose(score, 8.55801582, rtol=1e-5)  # should be the same everywhere
        ref_align = numpy.array(
            [
                [[1.0, 0.0, 0.0, 0.0, 0.0]],
                [[0.33333316, 0.66666663, 0.0, 0.0, 0.0]],
                [[0.06666669, 0.53333354, 0.40000018, 0.0, 0.0]],
                [[0.0, 0.20000014, 0.60000014, 0.19999999, 0.0]],
                [[0.0, 0.0, 0.39999962, 0.53333312, 0.06666663]],
                [[0.0, 0.0, 0.0, 0.66666633, 0.33333316]],
                [[0.0, 0.0, 0.0, 0.0, 0.99999982]],
            ],
            dtype=numpy.float32,
        )
        assert ref_align.shape == (seq_len, 1, n_classes)
        ref_align = numpy.tile(ref_align, (1, n_batch, 1))
        assert ref_align.shape == bw.shape
        # print("Reference alignment:")
        # print(repr(ref_align))
        print("mean square diff:", numpy.mean(numpy.square(ref_align - bw)))
        print("max square diff:", numpy.max(numpy.square(ref_align - bw)))
        assert_allclose(ref_align, bw, rtol=1e-5)
    print("Done.")


def _log_softmax(x, axis=-1):
    assert isinstance(x, numpy.ndarray)
    xdev = x - x.max(axis=axis, keepdims=True)
    lsm = xdev - numpy.log(numpy.sum(numpy.exp(xdev), axis=axis, keepdims=True))
    return lsm


def check_ctc_fsa(targets, target_seq_lens, n_classes, with_native_fsa=False, label_loop=True):
    """
    :param numpy.ndarray targets:
    :param numpy.ndarray target_seq_lens:
    :param int n_classes:
    :param bool with_native_fsa:
    :param bool label_loop: ctc_merge_repeated in tf.nn.ctc_loss
    :return: nothing, just checks
    """
    n_batch, n_target_time = targets.shape
    assert n_batch == len(target_seq_lens) and n_target_time == max(target_seq_lens)
    n_time = n_target_time * 3
    # am_scores are logits, unnormalized, i.e. the values before softmax.
    am_scores = numpy.random.RandomState(42).normal(size=(n_time, n_batch, n_classes)).astype("float32")
    # am_scores = numpy.zeros((n_time, n_batch, n_classes), dtype="float32")
    int_idx = numpy.zeros((n_time, n_batch), dtype="int32")
    seq_lens = numpy.array(
        [
            n_time,
            max(n_time - 4, (target_seq_lens[1] if (len(target_seq_lens) >= 2) else 0) + 1, 1),
            max(n_time - 5, 1),
            max(n_time - 5, 1),
        ],
        dtype="int32",
    )[:n_batch]
    for t in range(n_time):
        int_idx[t] = t < seq_lens
    float_idx = int_idx.astype("float32")
    blank_idx = n_classes - 1

    targets_th = torch.tensor(targets)
    targets_seq_lens_th = torch.tensor(target_seq_lens)

    if label_loop:
        fsa = get_ctc_fsa_fast_bw_np(targets=targets, seq_lens=target_seq_lens, blank_idx=blank_idx)
        assert fsa.start_end_states.shape == (2, len(target_seq_lens))
        edges = fsa.edges.astype("int32")
        weights = fsa.weights.astype("float32")
        start_end_states = fsa.start_end_states.astype("int32")
        if with_native_fsa:
            print("python edges:")
            print(edges)
            print("python start_end_states:")
            print(start_end_states)
    else:
        native_edges_th, native_weights_th, native_start_end_states_th = get_ctc_fsa_fast_bw(
            targets=targets_th, seq_lens=targets_seq_lens_th, blank_idx=blank_idx, label_loop=label_loop
        )
        edges, weights, start_end_states = (
            native_edges_th.numpy(),
            native_weights_th.numpy(),
            native_start_end_states_th.numpy(),
        )

    fwdbwd, obs_scores = py_baum_welch(
        am_scores=-_log_softmax(am_scores),
        float_idx=float_idx,
        edges=edges,
        weights=weights,
        start_end_states=start_end_states,
    )
    fwdbwd = numpy.exp(-fwdbwd)  # -log space -> prob space
    print(fwdbwd)
    print(obs_scores)

    if with_native_fsa:
        native_edges_th, native_weights_th, native_start_end_states_th = get_ctc_fsa_fast_bw(
            targets=targets_th, seq_lens=targets_seq_lens_th, blank_idx=blank_idx
        )
        native_edges, native_weights, native_start_end_states = (
            native_edges_th.numpy(),
            native_weights_th.numpy(),
            native_start_end_states_th.numpy(),
        )
        # Note: The native FSA vs the Python FSA are not exactly identical
        # (they just should be equivalent; although almost identical).
        # We introduce a dummy state (last before end state), and some dummy edges.
        print("native edges:")
        print(native_edges)
        print("native_start_end_states:")
        print(native_start_end_states)

        native_fwdbwd, native_obs_scores = py_baum_welch(
            am_scores=-_log_softmax(am_scores),
            float_idx=float_idx,
            edges=native_edges,
            weights=native_weights,
            start_end_states=native_start_end_states,
        )
        native_fwdbwd = numpy.exp(-native_fwdbwd)  # -log space -> prob space
        print(native_fwdbwd)
        print(native_obs_scores)
        for b in range(n_batch):
            for t in range(seq_lens[b]):
                numpy.testing.assert_almost_equal(fwdbwd[t, b], native_fwdbwd[t, b], decimal=5)
        for b in range(n_batch):
            numpy.testing.assert_almost_equal(obs_scores[b], native_obs_scores[b], decimal=5)
        fwdbwd = native_fwdbwd
        obs_scores = native_obs_scores

    if not label_loop:
        raise SkipTest("skip ref ctc_loss for label_loop=False")

    am_scores_th = torch.tensor(am_scores)
    seq_lens_th = torch.tensor(seq_lens)
    am_scores_th.requires_grad_()
    log_probs_th = torch.log_softmax(am_scores_th, dim=-1)
    ref_ctc_loss_th = torch.nn.functional.ctc_loss(
        log_probs=log_probs_th,
        input_lengths=seq_lens_th,
        targets=targets_th,
        target_lengths=targets_seq_lens_th,
        blank=blank_idx,
        reduction="none",
    )
    ref_ctc_loss_th.sum().backward()
    # See grad definition of CTCLoss.
    # The op will calculate the gradient w.r.t. the logits (log softmax).
    # I.e. with y = softmax(z), this is \partial loss / \partial z = y - soft_align.
    # Also see CtcLoss.get_soft_alignment.
    ref_ctc_loss_grad_th = am_scores_th.grad  # time major, i.e. (time, batch, dim)
    y_th = torch.softmax(am_scores_th, dim=-1)  # (time, batch, dim)
    soft_align_th = y_th - ref_ctc_loss_grad_th
    ref_fwdbwd, ref_obs_score = (soft_align_th.detach().numpy(), ref_ctc_loss_th.detach().numpy())
    print(ref_fwdbwd)
    print(ref_obs_score)

    for b in range(n_batch):
        for t in range(seq_lens[b]):
            numpy.testing.assert_almost_equal(fwdbwd[t, b], ref_fwdbwd[t, b], decimal=5)
    for b in range(n_batch):
        numpy.testing.assert_almost_equal(obs_scores[0, b], ref_obs_score[b], decimal=5)


def test_ctc_fsa_batch3_len6_c8():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[1, 3, 4, 2, 1, 0], [2, 6, 3, 4, 0, 0], [0, 3, 2, 0, 0, 0]], dtype="int32")
    target_seq_lens = numpy.array([6, 4, 3], dtype="int32")
    n_classes = 8  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes)


def test_ctc_fsa_batch1_len2():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0, 1]], dtype="int32")
    target_seq_lens = numpy.array([2], dtype="int32")
    n_classes = 3  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes)


def test_ctc_fsa_batch1_len1():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0]], dtype="int32")
    target_seq_lens = numpy.array([1], dtype="int32")
    n_classes = 2  # +1 because of blank

    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes)


def test_ctc_fsa_batch3_len6_c8_native():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[1, 3, 4, 2, 1, 0], [2, 6, 3, 4, 0, 0], [0, 3, 2, 0, 0, 0]], dtype="int32")
    target_seq_lens = numpy.array([6, 4, 3], dtype="int32")
    n_classes = 8  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch4_len6_c8_native():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array(
        [[1, 2, 4, 4, 1, 0], [2, 6, 3, 4, 0, 0], [3, 3, 0, 0, 0, 0], [5, 0, 0, 0, 0, 0]], dtype="int32"
    )
    target_seq_lens = numpy.array([6, 4, 2, 1], dtype="int32")
    n_classes = 8  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch4_len6_c8_native_no_loop():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    No label loop (ctc_merge_repeated False) is equivalent to the Recurrent Neural Aligner (RNA) topology.
    """
    targets = numpy.array(
        [[1, 2, 4, 4, 1, 0], [2, 6, 3, 4, 0, 0], [3, 3, 0, 0, 0, 0], [5, 0, 0, 0, 0, 0]], dtype="int32"
    )
    target_seq_lens = numpy.array([6, 4, 2, 1], dtype="int32")
    n_classes = 8  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, label_loop=False)


def test_ctc_fsa_batch2_len2a():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0, 1], [1, 0]], dtype="int32")
    target_seq_lens = numpy.array([2, 1], dtype="int32")
    n_classes = 3  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch2_len2():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0, 1], [0, 0]], dtype="int32")
    target_seq_lens = numpy.array([2, 2], dtype="int32")
    n_classes = 3  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch2_len1():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0], [1]], dtype="int32")
    target_seq_lens = numpy.array([1, 1], dtype="int32")
    n_classes = 3  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch1_len2rep_native():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0, 0]], dtype="int32")
    target_seq_lens = numpy.array([2], dtype="int32")
    n_classes = 2  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch1_len2_native():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0, 1]], dtype="int32")
    target_seq_lens = numpy.array([2], dtype="int32")
    n_classes = 3  # +1 because of blank
    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_ctc_fsa_batch1_len1_native():
    """
    This (:func:`Fsa.get_ctc_fsa_fast_bw`) is used by :func:`ctc_loss`.
    """
    targets = numpy.array([[0]], dtype="int32")
    target_seq_lens = numpy.array([1], dtype="int32")
    n_classes = 2  # +1 because of blank

    check_ctc_fsa(targets=targets, target_seq_lens=target_seq_lens, n_classes=n_classes, with_native_fsa=True)


def test_fast_viterbi(device: torch.device = torch.device("cpu")):
    n_batch = 3
    seq_len = 7
    n_classes = 5

    fsa = FastBwFsaShared()
    for i in range(n_classes):
        fsa.add_edge(i, i + 1, emission_idx=i)  # fwd
        fsa.add_edge(i + 1, i + 1, emission_idx=i)  # loop
    assert n_classes <= seq_len
    fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
    edges = fast_bw_fsa.edges
    weights = fast_bw_fsa.weights
    start_end_states = fast_bw_fsa.start_end_states
    am_scores = numpy.eye(n_classes, n_classes, dtype="float32")  # (dim,dim)

    am_scores = scipy.ndimage.zoom(am_scores, zoom=(float(seq_len) / n_classes, 1), order=1, prefilter=False)
    assert am_scores.shape == (seq_len, n_classes)
    am_scores = am_scores[:, None]
    am_scores = am_scores + numpy.zeros((seq_len, n_batch, n_classes), dtype="float32")
    print(am_scores[:, 0])
    # am_scores = numpy.ones((seq_len, n_batch, n_classes), dtype="float32") * numpy.float32(1.0 / n_classes)
    am_scores = numpy.log(am_scores)  # in +log space
    print("Construct call...")
    alignment, obs_scores = fast_viterbi(
        am_scores=torch.tensor(am_scores, device=device),
        am_seq_len=torch.tensor(numpy.array([seq_len] * n_batch, dtype="int32"), device=device),
        edges=torch.tensor(edges, device=device),
        weights=torch.tensor(weights, device=device),
        start_end_states=torch.tensor(start_end_states, device=device),
    )
    alignment, obs_scores = alignment.cpu().numpy(), obs_scores.cpu().numpy()
    print("Done.")
    print("score:")
    print(repr(obs_scores))
    assert obs_scores.shape == (n_batch,)
    print("Hard alignment:")
    print(repr(alignment))
    assert alignment.shape == (seq_len, n_batch)
    if seq_len == n_classes:
        print("Extra check identity...")
        for i in range(n_batch):
            for t in range(seq_len):
                assert alignment[t, i] == t
    if seq_len == 7 and n_classes == 5:
        print("Extra check ref_align (7,5)...")
        assert_allclose(obs_scores, -1.6218603, rtol=1e-5)  # should be the same everywhere
        for i in range(n_batch):
            assert alignment[:, i].tolist() == [0, 1, 1, 2, 3, 3, 4]
    print("Done.")


def test_fast_viterbi_cuda():
    if not torch.cuda.is_available():
        raise SkipTest("CUDA not available")
    test_fast_viterbi(device=torch.device("cuda"))


def test_fast_viterbi_rnd():
    n_batch = 4
    seq_len = 23
    n_classes = 5

    fsa = FastBwFsaShared()
    for i in range(n_classes):
        fsa.add_edge(i, i + 1, emission_idx=i)  # fwd
        fsa.add_edge(i + 1, i + 1, emission_idx=i)  # loop
    assert n_classes <= seq_len
    fast_bw_fsa = fsa.get_fast_bw_fsa(n_batch=n_batch)
    edges = fast_bw_fsa.edges
    weights = fast_bw_fsa.weights
    start_end_states = fast_bw_fsa.start_end_states
    am_scores = numpy.random.RandomState(42).normal(size=(seq_len, n_batch, n_classes)).astype("float32")
    am_seq_len = numpy.array([seq_len] * n_batch, dtype="int32")
    am_seq_len[1] -= 1
    am_seq_len[-1] -= 2
    am_seq_len[-2] = max(n_classes - 1, 1)  # no path possible
    ref_alignment, ref_scores = py_viterbi(
        am_scores=am_scores, am_seq_len=am_seq_len, edges=edges, weights=weights, start_end_states=start_end_states
    )
    print("ref score:")
    print(repr(ref_scores))
    assert ref_scores.shape == (n_batch,)
    print("ref hard alignment:")
    print(repr(ref_alignment))
    assert ref_alignment.shape == (seq_len, n_batch)
    print("Construct fast_viterbi call...")
    alignment, scores = fast_viterbi(
        am_scores=torch.tensor(am_scores),
        am_seq_len=torch.tensor(am_seq_len),
        edges=torch.tensor(edges),
        weights=torch.tensor(weights),
        start_end_states=torch.tensor(start_end_states),
    )
    alignment, scores = alignment.numpy(), scores.numpy()
    print("Done.")
    print("score:")
    print(repr(scores))
    assert scores.shape == (n_batch,)
    print("Hard alignment:")
    print(repr(alignment))
    assert alignment.shape == (seq_len, n_batch)
    assert_allclose(scores, ref_scores, rtol=1e-5)
    assert_allclose(alignment, ref_alignment, rtol=1e-5)
    print("Done.")


def test_edit_distance():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 13
    n_b_max_len = 11
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    a_len_np = rnd.randint(1, n_a_max_len + 1, size=(n_batch,), dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    # Likely some high error. So make some explicit examples.
    expected_results: List[Optional[int]] = [None] * n_batch
    i = 0
    # One insertion/deletion.
    a_np[i, :1] = [1]
    a_len_np[i] = 1
    b_len_np[i] = 0
    expected_results[i] = 1
    i += 1
    # One deletion.
    a_np[i, :2] = [1, 2]
    b_np[i, :1] = [1]
    a_len_np[i] = 2
    b_len_np[i] = 1
    expected_results[i] = 1
    i += 1
    # One substitution + deletion.
    a_np[i, :2] = [1, 2]
    b_np[i, :1] = [3]
    a_len_np[i] = 2
    b_len_np[i] = 1
    expected_results[i] = 2
    i += 1
    # One substitution error.
    a_np[i, :4] = [1, 2, 3, 4]
    b_np[i, :4] = [1, 2, 4, 4]
    a_len_np[i] = 4
    b_len_np[i] = 4
    expected_results[i] = 1
    i += 1
    # One deletion error.
    a_np[i, :6] = [1, 2, 3, 3, 4, 5]
    b_np[i, :5] = [1, 2, 3, 4, 5]
    a_len_np[i] = 6
    b_len_np[i] = 5
    expected_results[i] = 1
    i += 1
    # One insertion error.
    a_np[i, :6] = [1, 2, 3, 4, 5, 6]
    b_np[i, :7] = [1, 2, 3, 4, 4, 5, 6]
    a_len_np[i] = 6
    b_len_np[i] = 7
    expected_results[i] = 1
    i += 1
    # Same.
    a_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    b_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    a_len_np[i] = 11
    b_len_np[i] = 11
    expected_results[i] = 0
    i += 1
    # Both full length. Error should be 2.
    a_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0, 0, 0]
    b_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0]
    a_len_np[i] = n_a_max_len
    b_len_np[i] = n_b_max_len
    expected_results[i] = 2
    i += 1
    assert n_batch - i >= 5  # still some random left
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)

    for i in range(n_batch):
        print("testing batch", i, "/", n_batch)
        _a = a[i : i + 1, : a_len_np[i]]
        _a_len = a_len[i : i + 1]
        _b = b[i : i + 1, : b_len_np[i]]
        _b_len = b_len[i : i + 1]
        print("seq a:", a_np[i, : a_len_np[i]])
        print("seq b:", b_np[i, : b_len_np[i]])
        ref_edit_dist_np = edit_distance_ref_np(_a.numpy(), _a_len.numpy(), _b.numpy(), _b_len.numpy())
        native_edit_dist_np = edit_distance(_a, _a_len, _b, _b_len).numpy()
        assert isinstance(ref_edit_dist_np, numpy.ndarray)
        assert isinstance(native_edit_dist_np, numpy.ndarray)
        print("Ref edit dist:", ref_edit_dist_np)
        print("Native edit dist:", native_edit_dist_np)
        print("Expected edit dist:", expected_results[i])
        assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (1,)
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[0] == native_edit_dist_np[0]
        else:
            assert ref_edit_dist_np[0] == native_edit_dist_np[0]
        print("swapped:")
        ref_edit_dist_np = edit_distance_ref_np(_b.numpy(), _b_len.numpy(), _a.numpy(), _a_len.numpy())
        native_edit_dist_np = edit_distance(_b, _b_len, _a, _a_len).numpy()
        assert isinstance(ref_edit_dist_np, numpy.ndarray)
        assert isinstance(native_edit_dist_np, numpy.ndarray)
        print("Ref edit dist:", ref_edit_dist_np)
        print("Native edit dist:", native_edit_dist_np)
        print("Expected edit dist:", expected_results[i])
        assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (1,)
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[0] == native_edit_dist_np[0]
        else:
            assert ref_edit_dist_np[0] == native_edit_dist_np[0]
        print()

    print("Now the whole batch.")
    ref_edit_dist_np = edit_distance_ref_np(a.numpy(), a_len.numpy(), b.numpy(), b_len.numpy())
    native_edit_dist_np = edit_distance(a, a_len, b, b_len).numpy()
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]
    print()

    print("Now the whole batch, flipped.")
    ref_edit_dist_np = edit_distance_ref_np(b.numpy(), b_len.numpy(), a.numpy(), a_len.numpy())
    native_edit_dist_np = edit_distance(b, b_len, a, a_len).numpy()
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]


def test_edit_distance_2():
    from returnn.tensor import Tensor, Dim
    import numpy
    import torch
    from collections import namedtuple
    import itertools
    from numpy_ref_edit_distance import edit_distance_ref

    # noinspection PyShadowingNames
    def _check_edit_distance(a: Tensor, a_spatial_dim: Dim, b: Tensor, b_spatial_dim: Dim):
        ref = edit_distance_ref(a, a_spatial_dim, b, b_spatial_dim).raw_tensor
        res = edit_distance(a.raw_tensor, a_spatial_dim.dyn_size, b.raw_tensor, b_spatial_dim.dyn_size)
        assert res.shape == ref.shape == a_spatial_dim.dyn_size.shape == b_spatial_dim.dyn_size.shape
        assert len(ref.shape) == 1
        print("ref:", ref, "res:", res)
        batch_size = ref.shape[0]
        for i in range(batch_size):
            assert res[i] == ref[i], (
                f"batch idx i={i}, a[i]={a.raw_tensor[i]} len {a_spatial_dim.dyn_size[i]},"
                f" b[i]={b.raw_tensor[i]} len {b_spatial_dim.dyn_size[i]},"
                f" ref[i]={ref[i]}, res[i]={res[i]};\n"
                f" a={a.raw_tensor} lens {a_spatial_dim.dyn_size},"
                f" b={b.raw_tensor} lens {b_spatial_dim.dyn_size}"
            )
        assert (res.numpy() == ref).all()

    SizedTensor = namedtuple("SizedTensor", ["tensor", "seq_lens"])

    _SeqsB1 = [
        SizedTensor(torch.tensor([[1, 2, 3, 4]]), torch.tensor([4])),
        SizedTensor(torch.tensor([[1, 2, 3]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[1, 2, 4]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[1, 4]]), torch.tensor([2])),
        SizedTensor(torch.tensor([[5, 2, 4]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[]], dtype=torch.int64), torch.tensor([0])),
    ]

    for a, b in itertools.product(_SeqsB1, _SeqsB1):
        a: SizedTensor
        b: SizedTensor
        # noinspection PyShadowingNames
        batch_dim = Dim(1, name="batch")
        a_spatial_dim = Dim(Tensor("a_sizes", [batch_dim], dtype="int64", raw_tensor=a.seq_lens))
        b_spatial_dim = Dim(Tensor("b_sizes", [batch_dim], dtype="int64", raw_tensor=b.seq_lens))
        a_ = Tensor("a", [batch_dim, a_spatial_dim], dtype="int64", raw_tensor=a.tensor)
        b_ = Tensor("b", [batch_dim, b_spatial_dim], dtype="int64", raw_tensor=b.tensor)
        _check_edit_distance(a_, a_spatial_dim, b_, b_spatial_dim)

    rnd = numpy.random.RandomState(42)
    for a, b in itertools.product(_SeqsB1, _SeqsB1):
        batch_size = rnd.randint(2, 11)
        a_max_len = rnd.randint(a.seq_lens[0], a.seq_lens[0] + 5)
        b_max_len = rnd.randint(b.seq_lens[0], b.seq_lens[0] + 5)
        a_sizes = rnd.randint(0, a_max_len + 1, size=(batch_size,))
        b_sizes = rnd.randint(0, b_max_len + 1, size=(batch_size,))
        a_sizes[0] = a.seq_lens[0]
        b_sizes[0] = b.seq_lens[0]
        a_max_len = max(a_sizes)
        b_max_len = max(b_sizes)
        a_values = rnd.randint(0, 10, (batch_size, a_max_len))
        b_values = rnd.randint(0, 10, (batch_size, b_max_len))
        a_values[0, : a.seq_lens[0]] = a.tensor[0, : a.seq_lens[0]]
        b_values[0, : b.seq_lens[0]] = b.tensor[0, : b.seq_lens[0]]
        a_sizes = torch.tensor(a_sizes, dtype=torch.int32)
        b_sizes = torch.tensor(b_sizes, dtype=torch.int32)

        # noinspection PyShadowingNames
        batch_dim = Dim(batch_size, name="batch")
        a_spatial_dim = Dim(Tensor("a_sizes", [batch_dim], dtype="int32", raw_tensor=a_sizes))
        b_spatial_dim = Dim(Tensor("b_sizes", [batch_dim], dtype="int32", raw_tensor=b_sizes))
        a_ = Tensor("a", [batch_dim, a_spatial_dim], dtype="int64", raw_tensor=torch.tensor(a_values))
        b_ = Tensor("b", [batch_dim, b_spatial_dim], dtype="int64", raw_tensor=torch.tensor(b_values))
        _check_edit_distance(a_, a_spatial_dim, b_, b_spatial_dim)


def test_edit_distance_ref_np_b1():
    assert edit_distance_ref_np_b1([1], []) == 1
    assert edit_distance_ref_np_b1([1, 2], [1]) == 1
    assert edit_distance_ref_np_b1([2, 2], [1]) == 2
    assert edit_distance_ref_np_b1([2, 1], [1]) == 1
    assert edit_distance_ref_np_b1([2, 1], [1, 1]) == 1
    assert edit_distance_ref_np_b1([2, 1], [1, 1, 1]) == 2
    assert edit_distance_ref_np_b1([2, 1], [2, 1, 1]) == 1


def _naive_optimal_completion_edit_distance(
    a: Union[Sequence[int], numpy.ndarray], b: Union[Sequence[int], numpy.ndarray]
) -> int:
    distances = [edit_distance_ref_np_b1(a, b[:n]) for n in range(len(b) + 1)]
    return min(distances)


def test_optimal_completion_edit_distance():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 11
    n_b_max_len = 13
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    a_len_np = rnd.randint(1, n_a_max_len + 1, size=(n_batch,), dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    # Likely some high error. So make some explicit examples.
    expected_results = [None] * n_batch
    i = 0
    # One deletion.
    a_np[i, :1] = [1]
    a_len_np[i] = 1
    b_len_np[i] = 0
    expected_results[i] = 1
    i += 1
    # One optional insertion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [1, 2]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 0
    i += 1
    # One substitution or deletion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [3, 1]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 1
    i += 1
    # One substitution error.
    a_np[i, :4] = [1, 2, 3, 4]
    b_np[i, :4] = [1, 2, 4, 4]
    a_len_np[i] = 4
    b_len_np[i] = 4
    expected_results[i] = 1
    i += 1
    # One insertion error.
    a_np[i, :5] = [1, 2, 3, 4, 5]
    b_np[i, :6] = [1, 2, 3, 3, 4, 5]
    a_len_np[i] = 5
    b_len_np[i] = 6
    expected_results[i] = 1
    i += 1
    # Same.
    a_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    b_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    a_len_np[i] = 11
    b_len_np[i] = 11
    expected_results[i] = 0
    i += 1
    # Both full length.
    a_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0]
    b_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0, 0, 0]
    a_len_np[i] = n_a_max_len
    b_len_np[i] = n_b_max_len
    expected_results[i] = 0
    i += 1
    assert n_batch - i >= 5  # still some random left
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)
    for i in range(n_batch):
        print("testing batch", i, "/", n_batch)
        _a = a[i : i + 1, : a_len_np[i]]
        _a_len = a_len[i : i + 1]
        _b = b[i : i + 1, : b_len_np[i]]
        _b_len = b_len[i : i + 1]
        print("seq a:", a_np[i, : a_len_np[i]])
        print("seq b:", b_np[i, : b_len_np[i]])
        _native_edit_dist = optimal_completion_edit_distance(_a, _a_len, _b, _b_len)
        native_edit_dist_np = _native_edit_dist.numpy()
        ref_edit_dist_np = numpy.array(
            [_naive_optimal_completion_edit_distance(a_np[i, : a_len_np[i]], b_np[i, : b_len_np[i]])]
        )
        assert isinstance(ref_edit_dist_np, numpy.ndarray)
        assert isinstance(native_edit_dist_np, numpy.ndarray)
        print("Ref edit dist:", ref_edit_dist_np)
        print("Native edit dist:", native_edit_dist_np)
        print("Expected edit dist:", expected_results[i])
        assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (1,)
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[0] == native_edit_dist_np[0]
        else:
            assert ref_edit_dist_np[0] == native_edit_dist_np[0]
        print()

    print("Now the whole batch.")
    native_edit_dist = optimal_completion_edit_distance(a, a_len, b, b_len)
    native_edit_dist_np = native_edit_dist.numpy()
    ref_edit_dist_np = numpy.array(
        [
            _naive_optimal_completion_edit_distance(a_np[i, : a_len_np[i]], b_np[i, : b_len_np[i]])
            for i in range(n_batch)
        ]
    )
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]
    print()


def test_optimal_completion_edit_distance_per_successor():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 11
    n_b_max_len = 13
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    a_len_np = rnd.randint(1, n_a_max_len + 1, size=(n_batch,), dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    # Likely some high error. So make some explicit examples.
    expected_results = [None] * n_batch
    i = 0
    # One deletion.
    a_np[i, :1] = [1]
    a_len_np[i] = 1
    b_len_np[i] = 0
    expected_results[i] = 1
    i += 1
    # One optional insertion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [1, 2]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 0
    i += 1
    # One substitution or deletion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [3, 1]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 1
    i += 1
    # One substitution error.
    a_np[i, :4] = [1, 2, 3, 4]
    b_np[i, :4] = [1, 2, 4, 4]
    a_len_np[i] = 4
    b_len_np[i] = 4
    expected_results[i] = 1
    i += 1
    # One insertion error.
    a_np[i, :5] = [1, 2, 3, 4, 5]
    b_np[i, :6] = [1, 2, 3, 3, 4, 5]
    a_len_np[i] = 5
    b_len_np[i] = 6
    expected_results[i] = 1
    i += 1
    # Same.
    a_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 1, 3]
    b_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 1, 3]
    a_len_np[i] = 11
    b_len_np[i] = 11
    expected_results[i] = 0
    i += 1
    # Both full length.
    a_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0]
    b_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0, 0, 0]
    a_len_np[i] = n_a_max_len
    b_len_np[i] = n_b_max_len
    expected_results[i] = 0
    i += 1
    assert n_batch - i >= 5  # still some random left
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    assert all(a_len_np > 0)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)
    print("Now the whole batch.")
    # a_len - 1 such that we can do the check below.
    native_edit_dist = optimal_completion_edit_distance_per_successor(a, a_len - 1, b, b_len, num_classes)
    native_edit_dist_np = native_edit_dist.numpy()
    ref_edit_dist_np = numpy.array(
        [
            _naive_optimal_completion_edit_distance(a_np[i, : a_len_np[i]], b_np[i, : b_len_np[i]])
            for i in range(n_batch)
        ]
    )
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == (n_batch,)
    assert native_edit_dist_np.shape == (n_batch, num_classes)
    for i in range(n_batch):
        a_last = a_np[i, a_len_np[i] - 1]
        assert 0 <= a_last < num_classes
        native_res = native_edit_dist_np[i, a_last]
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_res
        else:
            assert ref_edit_dist_np[i] == native_res
        for j in range(num_classes):
            ref_res = _naive_optimal_completion_edit_distance(
                list(a_np[i, : a_len_np[i] - 1]) + [j], list(b_np[i, : b_len_np[i]])
            )
            native_res = native_edit_dist_np[i, j]
            assert ref_res == native_res
    print()


def test_next_edit_distance_row():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 13
    n_b_max_len = 11
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    a_len_np = rnd.randint(1, n_a_max_len + 1, size=(n_batch,), dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    # Likely some high error. So make some explicit examples.
    expected_results = [None] * n_batch
    i = 0
    # One insertion/deletion.
    a_np[i, :1] = [1]
    a_len_np[i] = 1
    b_len_np[i] = 0
    expected_results[i] = 1
    i += 1
    # One deletion.
    a_np[i, :2] = [1, 2]
    b_np[i, :1] = [1]
    a_len_np[i] = 2
    b_len_np[i] = 1
    expected_results[i] = 1
    i += 1
    # One substitution + deletion.
    a_np[i, :2] = [1, 2]
    b_np[i, :1] = [3]
    a_len_np[i] = 2
    b_len_np[i] = 1
    expected_results[i] = 2
    i += 1
    # One substitution error.
    a_np[i, :4] = [1, 2, 3, 4]
    b_np[i, :4] = [1, 2, 4, 4]
    a_len_np[i] = 4
    b_len_np[i] = 4
    expected_results[i] = 1
    i += 1
    # One deletion error.
    a_np[i, :6] = [1, 2, 3, 3, 4, 5]
    b_np[i, :5] = [1, 2, 3, 4, 5]
    a_len_np[i] = 6
    b_len_np[i] = 5
    expected_results[i] = 1
    i += 1
    # One insertion error.
    a_np[i, :6] = [1, 2, 3, 4, 5, 6]
    b_np[i, :7] = [1, 2, 3, 4, 4, 5, 6]
    a_len_np[i] = 6
    b_len_np[i] = 7
    expected_results[i] = 1
    i += 1
    # Same.
    a_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    b_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    a_len_np[i] = 11
    b_len_np[i] = 11
    expected_results[i] = 0
    i += 1
    # Both full length. Error should be 2.
    a_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0, 0, 0]
    b_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0]
    a_len_np[i] = n_a_max_len
    b_len_np[i] = n_b_max_len
    expected_results[i] = 2
    i += 1
    assert n_batch - i >= 5  # still some random left
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)

    for i in range(n_batch):
        print("testing batch", i, "/", n_batch)
        _a = a[i : i + 1, : a_len_np[i]]
        _a_len = a_len[i : i + 1]
        _b = b[i : i + 1, : b_len_np[i]]
        _b_len = b_len[i : i + 1]
        print("seq a:", a_np[i, : a_len_np[i]])
        print("seq b:", b_np[i, : b_len_np[i]])
        ref_edit_dist_np = edit_distance_ref_np(_a.numpy(), _a_len.numpy(), _b.numpy(), _b_len.numpy())
        native_edit_dist_np = edit_distance_via_next_edit_distance_row(_a, _a_len, _b, _b_len).numpy()
        assert isinstance(ref_edit_dist_np, numpy.ndarray)
        assert isinstance(native_edit_dist_np, numpy.ndarray)
        print("Ref edit dist:", ref_edit_dist_np)
        print("Native edit dist:", native_edit_dist_np)
        print("Expected edit dist:", expected_results[i])
        assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (1,)
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[0] == native_edit_dist_np[0]
        else:
            assert ref_edit_dist_np[0] == native_edit_dist_np[0]
        print("swapped:")
        ref_edit_dist_np = edit_distance_ref_np(_b.numpy(), _b_len.numpy(), _a.numpy(), _a_len.numpy())
        native_edit_dist_np = edit_distance_via_next_edit_distance_row(_b, _b_len, _a, _a_len).numpy()
        assert isinstance(ref_edit_dist_np, numpy.ndarray)
        assert isinstance(native_edit_dist_np, numpy.ndarray)
        print("Ref edit dist:", ref_edit_dist_np)
        print("Native edit dist:", native_edit_dist_np)
        print("Expected edit dist:", expected_results[i])
        assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (1,)
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[0] == native_edit_dist_np[0]
        else:
            assert ref_edit_dist_np[0] == native_edit_dist_np[0]
        print()

    print("Now the whole batch.")
    ref_edit_dist_np = edit_distance_ref_np(a.numpy(), a_len.numpy(), b.numpy(), b_len.numpy())
    native_edit_dist_np = edit_distance_via_next_edit_distance_row(a, a_len, b, b_len).numpy()
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]
    print()

    print("Now the whole batch, flipped.")
    ref_edit_dist_np = edit_distance_ref_np(b.numpy(), b_len.numpy(), a.numpy(), a_len.numpy())
    native_edit_dist_np = edit_distance_via_next_edit_distance_row(b, b_len, a, a_len).numpy()
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]


def test_next_edit_distance_row_optimal_completion():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 11
    n_b_max_len = 13
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    a_len_np = rnd.randint(1, n_a_max_len + 1, size=(n_batch,), dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    # Likely some high error. So make some explicit examples.
    expected_results = [None] * n_batch
    i = 0
    # One deletion.
    a_np[i, :1] = [1]
    a_len_np[i] = 1
    b_len_np[i] = 0
    expected_results[i] = 1
    i += 1
    # One optional insertion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [1, 2]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 0
    i += 1
    # One substitution or deletion.
    a_np[i, :1] = [1]
    b_np[i, :2] = [3, 1]
    a_len_np[i] = 1
    b_len_np[i] = 2
    expected_results[i] = 1
    i += 1
    # One substitution error.
    a_np[i, :4] = [1, 2, 3, 4]
    b_np[i, :4] = [1, 2, 4, 4]
    a_len_np[i] = 4
    b_len_np[i] = 4
    expected_results[i] = 1
    i += 1
    # One insertion error.
    a_np[i, :5] = [1, 2, 3, 4, 5]
    b_np[i, :6] = [1, 2, 3, 3, 4, 5]
    a_len_np[i] = 5
    b_len_np[i] = 6
    expected_results[i] = 1
    i += 1
    # Same.
    a_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    b_np[i, :11] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 11]
    a_len_np[i] = 11
    b_len_np[i] = 11
    expected_results[i] = 0
    i += 1
    # Both full length.
    a_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0]
    b_np[i] = [2, 2, 4, 4, 6, 6, 8, 8, 9, 10, 0, 0, 0]
    a_len_np[i] = n_a_max_len
    b_len_np[i] = n_b_max_len
    expected_results[i] = 0
    i += 1
    assert n_batch - i >= 5  # still some random left
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)
    print("Now the whole batch.")
    native_edit_dist = edit_distance_via_next_edit_distance_row(a, a_len, b, b_len, optimal_completion=True)
    native_edit_dist_np = native_edit_dist.numpy()
    ref_edit_dist_np = numpy.array(
        [
            _naive_optimal_completion_edit_distance(a_np[i, : a_len_np[i]], b_np[i, : b_len_np[i]])
            for i in range(n_batch)
        ]
    )
    assert isinstance(ref_edit_dist_np, numpy.ndarray)
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Ref edit dist:", ref_edit_dist_np)
    print("Native edit dist:", native_edit_dist_np)
    print("Expected edit dist:", expected_results)
    assert ref_edit_dist_np.shape == native_edit_dist_np.shape == (n_batch,)
    for i in range(n_batch):
        if expected_results[i] is not None:
            assert expected_results[i] == ref_edit_dist_np[i] == native_edit_dist_np[i]
        else:
            assert ref_edit_dist_np[i] == native_edit_dist_np[i]
    print()


def test_next_edit_distance_reduce_optimal_completion():
    rnd = numpy.random.RandomState(42)
    n_batch = 15
    n_a_max_len = 7
    n_b_max_len = 13
    num_classes = 10
    a_np = rnd.randint(0, num_classes, size=(n_batch, n_a_max_len), dtype="int32")
    b_np = rnd.randint(0, num_classes, size=(n_batch, n_b_max_len), dtype="int32")
    # Test only for full length seqs in a.
    a_len_np = numpy.array([n_a_max_len] * n_batch, dtype="int32")
    b_len_np = rnd.randint(1, n_b_max_len + 1, size=(n_batch,), dtype="int32")
    a = torch.tensor(a_np)
    b = torch.tensor(b_np)
    assert all(a_len_np > 0)
    a_len = torch.tensor(a_len_np)
    b_len = torch.tensor(b_len_np)
    print("Now the whole batch.")
    # a_len - 1 such that we can do the check below.
    native_edit_dist = optimal_completion_edit_distance_per_successor_via_next_edit_distance(
        a, a_len, b, b_len, num_classes
    )
    native_edit_dist_np = native_edit_dist.numpy()
    assert isinstance(native_edit_dist_np, numpy.ndarray)
    print("Native edit dist:", native_edit_dist_np)
    assert native_edit_dist_np.shape == (n_batch, num_classes)
    for i in range(n_batch):
        for j in range(num_classes):
            ref_res = _naive_optimal_completion_edit_distance(
                list(a_np[i, : a_len_np[i]]) + [j], b_np[i, : b_len_np[i]]
            )
            native_res = native_edit_dist_np[i, j]
            assert ref_res == native_res
    print()



"""
Reference implementation of edit distance.
"""

from typing import Union, Sequence
import numpy
from returnn.tensor import Tensor, Dim


def edit_distance_ref(a: Tensor, a_spatial_dim: Dim, b: Tensor, b_spatial_dim: Dim) -> Tensor:
    """
    Reference implementation for edit distance.
    """
    batch_dim = a.dims[0]
    assert a.dims == (batch_dim, a_spatial_dim) and b.dims == (batch_dim, b_spatial_dim)
    res = []
    for i in range(batch_dim.dimension):
        assert a_spatial_dim.dyn_size[i] <= a.raw_tensor.size(1)
        assert b_spatial_dim.dyn_size[i] <= b.raw_tensor.size(1)
        res.append(
            edit_distance_ref_np_b1(
                a.raw_tensor[i, : a_spatial_dim.dyn_size[i]], b.raw_tensor[i, : b_spatial_dim.dyn_size[i]]
            )
        )
    return Tensor("edit_dist", dims=[batch_dim], dtype="int32", raw_tensor=numpy.array(res, dtype=numpy.int32))


def edit_distance_ref_np(
    a: numpy.ndarray, a_seq_lens: numpy.ndarray, b: numpy.ndarray, b_seq_lens: numpy.ndarray
) -> numpy.ndarray:
    """
    Reference implementation for edit distance.
    """
    batch_size = a.shape[0]
    assert a.shape == (batch_size, a_seq_lens.max()) and b.shape == (batch_size, b_seq_lens.max())
    res = []
    for i in range(batch_size):
        assert a_seq_lens[i] <= a.shape[1]
        assert b_seq_lens[i] <= b.shape[1]
        res.append(edit_distance_ref_np_b1(a[i, : a_seq_lens[i]], b[i, : b_seq_lens[i]]))
    return numpy.array(res, dtype=numpy.int32)


def edit_distance_ref_np_b1(a: Union[Sequence[int], numpy.ndarray], b: Union[Sequence[int], numpy.ndarray]) -> int:
    """
    Reference implementation for edit distance, single batch.
    """
    n = len(a) + 1
    m = len(b) + 1
    d = numpy.zeros((n, m), dtype=numpy.int32)
    for i in range(n):
        d[i, 0] = i
    for j in range(m):
        d[0, j] = j
    for j in range(1, m):
        for i in range(1, n):
            if a[i - 1] == b[j - 1]:
                d[i, j] = d[i - 1, j - 1]
            else:
                d[i, j] = min(
                    d[i - 1, j] + 1,  # deletion
                    d[i, j - 1] + 1,  # insertion
                    d[i - 1, j - 1] + 1,  # substitution
                )
    return int(d[n - 1, m - 1])



"""
Native ops for Torch, similar to :mod:`returnn.tf.native_op`.
"""

from __future__ import annotations
from typing import Optional, Any, Tuple, Dict
import os
import sys
from textwrap import dedent
from threading import RLock

import torch

from returnn import native_op
from .native_op_code_compiler import OpCodeCompiler


_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
_base_dir = os.path.realpath(_base_dir)  # Make canonical path-name.


class OpDescription(native_op.NativeOpBaseMixin):
    """
    Meta-info about an op, used by :class:`OpMaker`.
    """

    @classmethod
    def from_gen_base(cls, gen_base):
        """
        :param returnn.native_op.NativeOpGenBase|type[returnn.native_op.NativeOpGenBase] gen_base:
        :rtype: OpDescription
        """
        name = gen_base.__name__
        assert gen_base.in_info is not None
        assert gen_base.out_info is not None
        assert gen_base.c_fw_code is not None
        return OpDescription(
            in_info=gen_base.in_info,
            out_info=gen_base.out_info,
            c_fw_code=gen_base.c_fw_code,
            c_bw_code=gen_base.c_bw_code,
            c_extra_support_code=gen_base.c_extra_support_code,
            cpu_support=gen_base.cpu_support,
            grad_input_map=gen_base.grad_input_map,
            name=name,
        )

    @property
    def is_grad_defined(self) -> bool:
        """
        :return: whether the gradient is defined
        """
        return bool(self.c_bw_code)

    def grad(self) -> Optional[OpDescription]:
        """
        :rtype: OpDescription|None
        """
        if not self.is_grad_defined:
            return None
        kwargs = self.kwargs_for_grad_op()
        return OpDescription(**kwargs)


class OpMaker:
    """
    https://docs.pytorch.org/tutorials/advanced/cpp_custom_ops.html
    https://docs.pytorch.org/cppdocs/
    """

    with_cuda: Optional[bool] = None
    global_lock = RLock()
    mod_cache = {}  # cache_key -> mod
    op_cache = {}  # cache_key -> op
    log_stream = sys.stdout

    def __init__(
        self,
        description: OpDescription,
        *,
        compiler_opts: Optional[Dict[str, str]] = None,
        with_cuda: Optional[bool] = None,
    ):
        """
        :param description:
        :param compiler_opts: passed on to OpCodeCompiler as kwargs
        :param with_cuda: override auto-detection of CUDA availability
        """
        if with_cuda is not None:
            self.with_cuda = with_cuda
        else:
            self._cls_init_with_cuda()
        self.description = description
        self.name = description.name
        self.compiler_opts = compiler_opts or {}

    @classmethod
    def _cls_init_with_cuda(cls):
        if cls.with_cuda is None:
            cls.with_cuda = torch.cuda.is_available()

    @property
    def op_name(self) -> str:
        """op name"""
        return self.name

    @property
    def cache_key(self) -> str:
        """cache key"""
        return self.name

    @property
    def support_native_op_cpp_filename(self) -> str:
        """
        :return: filename of NativeOp.cpp
        """
        support_native_op_cpp_filename = "%s/native_op.cpp" % _base_dir
        assert os.path.exists(support_native_op_cpp_filename)
        return support_native_op_cpp_filename

    def _make_code(self):
        # https://docs.pytorch.org/tutorials/advanced/cpp_custom_ops.html
        # https://docs.pytorch.org/cppdocs/

        # We also include NativeOp.cpp.

        # noinspection PyProtectedMember
        in_info, out_info, _ = native_op.NativeOpBaseMixin._resolve_want_inplace_dummy(
            in_info=self.description.in_info, out_info=self.description.out_info
        )

        # noinspection PyShadowingNames
        def map_name(v: Dict[str, Any], is_out: bool = False) -> str:
            """name"""
            name = v["name"].lower()
            if is_out:
                name = "_out_%s" % name
            else:
                name = "_in_%s" % name
            return name

        # noinspection PyShadowingNames,PyUnusedLocal
        def map_type(v: Dict[str, Any]) -> str:
            """dtype"""
            t = v.get("dtype", "float32")
            if t == "float32":
                return "at::kFloat"
            elif t == "int32":
                return "at::kInt"
            else:
                raise NotImplementedError("unsupported dtype %r" % t)

        def make_compute_code(*, cuda: bool = False) -> str:
            """compute code"""

            code_device_specific = ""
            if cuda:
                tensor_vars = [v for v in in_info if _schema_type_str(v) == "Tensor"]
                if tensor_vars:
                    code_device_specific += dedent(f"""\
                        at::Device _device = {map_name(tensor_vars[0])}.device();
                        at::OptionalDeviceGuard _device_guard(_device);
                        """)
                else:
                    code_device_specific += "at::Device _device = at::kCUDA;\n"

            code_forward_io = ""
            out_is_ref = {}  # output vars which are inplace, out_name -> in_idx
            # want_inplace: output-index which this input should operate on
            # Unlike the Theano variant, we always do it inplace,
            # so the user has to make a copy if this is not the intention.
            for in_idx, v in enumerate(in_info):
                out_idx = v.get("want_inplace", -1)
                if out_idx >= 0:
                    code_forward_io += dedent(f"""\
                        torch::Tensor {map_name(out_info[out_idx], is_out=True)} = {map_name(v)};  // inplace
                        """)
                    out_name = out_info[out_idx]["name"]
                    assert out_name not in out_is_ref
                    out_is_ref[out_name] = in_idx

            code_set_io = ""
            for in_idx, v in enumerate(in_info):
                if _schema_type_str(v) != "Tensor":
                    code_set_io += dedent(f"""\
                        torch::Tensor {map_name(v)}_tensor = torch::tensor({map_name(v)}, torch::dtype({map_type(v)}));
                        """)
                    continue  # scalar input
                ndim = len(v["shape"])
                code_set_io += dedent(f"""\
                    if({map_name(v)}.scalar_type() != {map_type(v)}) {{
                        {map_name(v)} = {map_name(v)}.to(torch::dtype({map_type(v)}));
                    }}
                    """)
                code_set_io += dedent(f"""\
                    TORCH_CHECK(
                        {map_name(v)}.dim() == {ndim},
                        "{v["name"]} shape ndim is not {ndim}, got shape ", {map_name(v)}.sizes());
                    """)
                for axis, d in enumerate(v["shape"]):
                    if isinstance(d, int):
                        code_set_io += dedent(f"""\
                            TORCH_CHECK(
                                {map_name(v)}.size({axis}) == {d},
                                "{v["name"]} shape[{axis}] != {d}, got shape ", {map_name(v)}.sizes());
                            """)

            for out_idx, v in enumerate(out_info):
                out_name = out_info[out_idx]["name"]
                if out_name in out_is_ref:  # is ref on input
                    pass
                else:  # no ref
                    cshape = "{%s}" % ", ".join(
                        [
                            str(dim) if isinstance(dim, int) else f"{map_name(in_info[dim[0]])}.size({dim[1]})"
                            for dim in v["shape"]
                        ]
                    )
                    code_set_io += dedent(f"""\
                        torch::Tensor {map_name(v, is_out=True)}
                            = torch::zeros({cshape}, torch::dtype({map_type(v)}){".device(_device)" if cuda else ""});
                        """)

            code_set_contiguous = ""
            for v in in_info:
                if v.get("need_contiguous", False) and _schema_type_str(v) == "Tensor":
                    code_set_contiguous += dedent(f"""\
                        if(!{map_name(v)}.is_contiguous()) {{
                            {map_name(v)} = {map_name(v)}.contiguous();
                        }}
                        """)

            # The user code uses inputs and outputs arrays.
            _code_wrap_io_input_vars_list = [
                f"&{map_name(v)}" + ("" if _schema_type_str(v) == "Tensor" else "_tensor") for v in in_info
            ]
            code_wrap_io = dedent(f"""\
                static const int n_inputs = {len(in_info)}, n_outputs = {len(out_info)};
                torch::Tensor* inputs[n_inputs] = {{
                    {", ".join(_code_wrap_io_input_vars_list)} }};
                torch::Tensor* _outputs_ptr[n_outputs] = {{
                    {", ".join(f"&{map_name(v, is_out=True)}" for v in out_info)} }};
                torch::Tensor** outputs[n_outputs] = {{
                    {", ".join(f"&_outputs_ptr[{i}]" for i in range(len(out_info)))} }};
                """)

            code_user = self.description.c_fw_code % {"fail": "assert(false);"}

            code_return = "return std::make_tuple(%s);\n" % ", ".join([map_name(v, is_out=True) for v in out_info])

            code_compute = "\n\n".join(
                [
                    code_device_specific,
                    code_forward_io,
                    code_set_io,
                    code_set_contiguous,
                    code_wrap_io,
                    code_user,
                    code_return,
                ]
            )

            return code_compute

        code_header = ""
        code_header += dedent("""\
            #include <torch/extension.h>
            #include <torch/types.h>
            #include <c10/core/CPUAllocator.h>
            """)
        if self.with_cuda:
            code_header += dedent("""\
              #include <cuda.h>
              #include <cuda_runtime.h>
              #include <math_constants.h>
              #include <ATen/cuda/CUDAContext.h>
              #include <c10/cuda/CUDACachingAllocator.h>
              """)

        def _schema_type_str(v: Dict[str, Any], *, c: bool = False) -> str:
            if v.get("host_memory", False):
                assert v["ndim"] == 0  # not supported otherwise...
                dtype = v.get("dtype", "float32")
                if dtype == "float32":
                    if c:
                        return "float32_t"
                    return "float"
                elif dtype == "int32":
                    if c:
                        return "int64_t"  # int8_t, int64_t and bool are supported as an integral argument type
                    return "int"
                else:
                    raise NotImplementedError("unsupported dtype %r" % dtype)
            if c:
                return "torch::Tensor"
            return "Tensor"

        # https://github.com/pytorch/pytorch/blob/main/aten/src/ATen/native/README.md#func
        func_schema_str = "(%s)" % ", ".join(
            f"{_schema_type_str(v)} {v['name']}" for v in in_info
        ) + " -> (%s)" % ", ".join(_schema_type_str(v) for v in out_info)

        code_header += dedent(
            f"""\
            #define _ns  // so _ns::something will use the root namespace
            #define TORCH 1
            #define CUDA 0
            #include "{self.support_native_op_cpp_filename}"

            TORCH_LIBRARY({self.op_name}, m) {{
                m.def("{self.op_name}{func_schema_str}");
            }}

            """
        )

        if self.description.cpu_support:
            # noinspection PyProtectedMember
            code_cpu_op = self.description._reduce_c_extra_support_code(self.description.c_extra_support_code)
            code_cpu_op += dedent(
                f"""\

                std::tuple<{", ".join(["torch::Tensor"] * len(out_info))}>
                {self.op_name}_cpu(
                    {", ".join(f"{_schema_type_str(v, c=True)} {map_name(v)}" for v in in_info)}
                ) {{
                """
            )
            code_cpu_op += make_compute_code()
            code_cpu_op += dedent(
                f"""\
                }}

                TORCH_LIBRARY_IMPL({self.op_name}, CPU, m) {{
                    m.impl("{self.op_name}", &{self.op_name}_cpu);
                }}
                """
            )
        else:
            code_cpu_op = ""

        if self.with_cuda:
            # noinspection PyProtectedMember
            code_cuda_op = dedent(f"""\
                namespace _cuda_impl {{

                    #ifdef _ns
                      #undef _ns
                      #define _ns _ns
                    #endif
                    namespace _ns = ::_cuda_impl;
                    #undef Ndarray_memcpy
                    #undef Ndarray_memset
                    #undef Ndarray_sgemm
                    #undef Ndarray_sgemv
                    #undef Ndarray_sgemm_batched
                    #undef DEF_KERNEL
                    #undef start_dev_kernel
                    #undef assert_cmp
                    #undef threadIdx
                    #undef blockIdx
                    #undef blockDim
                    #undef gridDim
                    #undef DEF_SHARED
                    #undef DEV_FUNC
                    #undef HANDLE_LAST_ERROR
                    #undef HOST_FUNC
                    #undef INF_F
                    #undef NAN_F
                    #undef elem_atomic_add
                    #undef elem_atomic_cas
                    #undef elem_atomic_min
                    #undef float_as_int
                    #undef int_as_float
                    #undef start_dev_kernel2
                    #undef CHECK_WITH_MSG

                    #undef CUDA
                    #define CUDA 1

                    #include "{self.support_native_op_cpp_filename}"

                    #undef CUDA  // name collision in Torch code below
                """)
            # noinspection PyProtectedMember
            code_cuda_op += self.description._reduce_c_extra_support_code(self.description.c_extra_support_code)
            code_cuda_op += dedent(f"""\

                    std::tuple<{", ".join(["torch::Tensor"] * len(out_info))}>
                    {self.op_name}_cuda(
                        {", ".join(f"{_schema_type_str(v, c=True)} {map_name(v)}" for v in in_info)}
                    ) {{
                """)
            code_cuda_op += make_compute_code(cuda=True)
            code_cuda_op += dedent(f"""\
                    }}

                    TORCH_LIBRARY_IMPL({self.op_name}, CUDA, m) {{
                        m.impl("{self.op_name}", &{self.op_name}_cuda);
                    }}
                }}  // namespace _cuda_impl
                """)
        else:
            code_cuda_op = ""

        return code_header + code_cpu_op + code_cuda_op

    def _make_mod(self):
        if self.cache_key in self.mod_cache:
            return self.mod_cache[self.cache_key]

        comp = OpCodeCompiler(
            base_name=self.name,
            code_version=self.description.code_version,
            code=self._make_code(),
            include_deps=[self.support_native_op_cpp_filename],
            use_cuda_if_available=self.with_cuda,
            log_stream=self.log_stream,
            **dict(self.compiler_opts),
        )
        mod = comp.load_module()
        mod._op_compiler = comp
        self.mod_cache[self.cache_key] = mod
        return mod

    def make_op(self):
        """
        :return: op
        """
        with self.global_lock:
            if self.cache_key in self.op_cache:
                return self.op_cache[self.cache_key]
            mod = self._make_mod()
            op = getattr(mod, self.op_name)
            op._op_maker = self
            op._op_module = mod
            self.op_cache[self.cache_key] = op

            if self.description.is_grad_defined:
                pass  # not implemented yet...

        return op


def make_op(cls, **kwargs):
    """
    :param type[returnn.native_op.NativeOpGenBase] cls:
    :param kwargs: passed to OpMaker
    :return: op
    :rtype: (torch.Tensor) -> tuple[torch.Tensor]
    """
    maker = OpMaker(OpDescription.from_gen_base(cls), **kwargs)
    return maker.make_op()


def ctc_loss(
    *,
    logits: torch.Tensor,
    logits_seq_lens: torch.Tensor,
    targets: torch.Tensor,
    targets_seq_lens: torch.Tensor,
    label_loop: bool = True,
    logits_time_major: bool = False,
    logits_normalize: bool = True,
    blank_index: int = -1,
    max_approx: bool = False,
) -> torch.Tensor:
    """
    Similar to :func:`tf.nn.ctc_loss`.
    We use our :func:`fast_baum_welch`.
    Also see :class:`FastBaumWelchLoss`.

    :param logits: (time,batch,dim) or (batch,time,dim). unnormalized (before softmax)
    :param logits_seq_lens: shape (batch,) of int32|int64
    :param logits_time_major:
    :param targets: batch-major, [batch,time]
    :param targets_seq_lens: (batch,)
    :param label_loop: (ctc_merge_repeated in tf.nn.ctc_loss)
    :param logits_normalize: apply log_softmax on logits (default).
      if False, you might also set grad_wrt_softmax_in=False
    :param blank_index: vocab index of the blank symbol
    :param max_approx: use max approximation (Viterbi) instead of full sum
    :return: loss, shape (batch,)
    """
    from .array_ import sequence_mask_time_major

    assert logits.ndim == 3
    dim = logits.shape[-1]
    if not logits_time_major:
        logits = torch.transpose(logits, 0, 1)  # (time,batch,dim)

    if blank_index < 0:
        blank_index += dim
    assert 0 <= blank_index < dim
    edges, weights, start_end_states = get_ctc_fsa_fast_bw(
        targets=targets, seq_lens=targets_seq_lens, blank_idx=blank_index, label_loop=label_loop
    )

    seq_mask = sequence_mask_time_major(logits_seq_lens)  # (time,batch), bool

    if max_approx:
        log_probs = torch.log_softmax(logits, dim=-1) if logits_normalize else logits  # (time,batch,dim)
        alignment, _ = fast_viterbi(
            am_scores=log_probs,
            am_seq_len=logits_seq_lens,
            edges=edges,
            weights=weights,
            start_end_states=start_end_states,
            mask_idx=blank_index,
        )
        # alignment is (time,batch)
        log_probs_ = torch.gather(log_probs, 2, alignment.unsqueeze(-1))  # (time,batch,1)
        log_probs_ = log_probs_.squeeze(-1)  # (time,batch)
        log_probs_ = torch.where(seq_mask, log_probs_, 0.0)
        loss = -torch.sum(log_probs_, dim=0)  # (batch,)
        return loss

    loss = _FastBaumWelchScoresAutogradFunc.apply(logits, logits_normalize, seq_mask, edges, weights, start_end_states)
    return loss


# noinspection PyMethodOverriding,PyAbstractClass,PyMissingOrEmptyDocstring
class _FastBaumWelchScoresAutogradFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        logits: torch.Tensor,
        logits_normalize: bool,
        seq_mask: torch.Tensor,
        edges: torch.Tensor,
        weights: torch.Tensor,
        start_end_states: torch.Tensor,
        state_buffer: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if logits_normalize:
            log_sm = torch.log_softmax(logits, dim=-1)  # (time,batch,dim)
        else:
            log_sm = logits
        fwdbwd, obs_scores = fast_baum_welch(
            am_scores=-log_sm,
            seq_mask=seq_mask,
            edges=edges,
            weights=weights,
            start_end_states=start_end_states,
            state_buffer=state_buffer,
        )
        loss = obs_scores[0]  # (batch,)
        ctx.grad_wrt_softmax_in = logits_normalize
        if logits_normalize:
            ctx.save_for_backward(log_sm, seq_mask, fwdbwd)
        else:
            ctx.save_for_backward(seq_mask, fwdbwd)
        return loss

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> Tuple[Optional[torch.Tensor], ...]:
        if ctx.grad_wrt_softmax_in:
            log_sm, seq_mask, fwdbwd = ctx.saved_tensors
        else:
            log_sm = None
            seq_mask, fwdbwd = ctx.saved_tensors
        bw = torch.exp(-fwdbwd)  # (time,batch,dim)
        if ctx.grad_wrt_softmax_in:
            grad_x = torch.exp(log_sm) - bw  # (time,batch,dim)
        else:
            grad_x = -bw  # (time,batch,dim)
        grad_x = torch.where(seq_mask[:, None, :], grad_x, 0.0)
        grad_x *= grad_output[None, :, None]
        return grad_x, None, None, None, None, None, None, None


def get_ctc_fsa_fast_bw(
    *, targets: torch.Tensor, seq_lens: torch.Tensor, blank_idx: int, label_loop: bool = True
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    See :class:`NativeOp.GetCtcFsaFastBwOp`.
    Generates a FSA with CTC topology. The output format is compatible to :func:`fast_baum_welch`.

    :param targets: shape (batch,time), int32
    :param seq_lens: shape (batch), int32
    :param blank_idx: vocab index of the blank symbol
    :param label_loop: True -> normal CTC; False -> RNA-like
    :return: edges, weights, start_end_states;
        edges is (4,num_edges), int32, edges of the graph (from,to,emission_idx,sequence_idx).
        weights is (num_edges,), float32. all zero.
        start_end_states is (2,batch), int32, (start,end) state idx in FSA.
    """
    assert targets.ndim == 2
    targets = targets.to(torch.int32)
    n_batch, n_time = targets.shape

    from .assert_ import assert_

    # The check on the seq lens is important
    # because invalid seq lens might not directly lead to an error here
    # but it might just return an invalid FSA.
    # An invalid FSA can however later cause a crash in the FastBaumWelchOp.
    assert_(seq_lens.max() == n_time, "get_ctc_fsa_fast_bw seq_lens invalid")

    n_edges = n_batch * (5 * (n_time - 1) + 10)  # see op documentation
    weights = torch.zeros((n_edges,), device=targets.device)
    maker = OpMaker(OpDescription.from_gen_base(native_op.GetCtcFsaFastBwOp))
    op = maker.make_op()
    edges, start_end_states = op(targets, seq_lens, blank_idx, weights, label_loop)

    return edges, weights, start_end_states


def fast_baum_welch(
    *,
    am_scores: torch.Tensor,
    seq_mask: torch.Tensor,
    edges: torch.Tensor,
    weights: torch.Tensor,
    start_end_states: torch.Tensor,
    state_buffer: Optional[torch.Tensor] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    :param am_scores: (time, batch, dim), in -log space
    :param seq_mask: (time, batch) -> 0 or 1 (index mask, via seq lens)
    :param edges: (4,num_edges), edges of the graph (from,to,emission_idx,sequence_idx)
    :param weights: (num_edges,), weights of the edges
    :param start_end_states: (2, batch), (start,end) state idx in automaton.
        there is only one single automaton.
    :param state_buffer: (2, num_states)
    :return: (fwdbwd, obs_scores), fwdbwd is (time, batch, dim), obs_scores is (time, batch), in -log space
    """
    from .assert_ import assert_

    # edges, weights, start_end_states, state_buffer = SprintAlignmentAutomataOp(self.sprint_opts)(self.network.tags)
    op = make_fast_baum_welch_op()
    float_idx = seq_mask.float()
    if state_buffer is None:
        last_state_idx = start_end_states[1].max()  # see get_automata_for_batch
        assert_(last_state_idx >= 0, "fast_baum_welch last_state_idx must be >= 0")
        state_buffer = torch.zeros((2, last_state_idx + 1))
    fwdbwd, obs_scores = op(am_scores, edges, weights, start_end_states, float_idx, state_buffer)  # noqa
    return fwdbwd, obs_scores


def make_fast_baum_welch_op(**kwargs):
    """
    :return: op
    :rtype: (torch.Tensor) -> tuple[torch.Tensor]
    """
    maker = OpMaker(OpDescription.from_gen_base(native_op.FastBaumWelchOp), **kwargs)
    return maker.make_op()


def fast_viterbi(
    *,
    am_scores: torch.Tensor,
    am_seq_len: torch.Tensor,
    edges: torch.Tensor,
    weights: torch.Tensor,
    start_end_states: torch.Tensor,
    mask_idx: int = 0,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    :param am_scores: (time, batch, dim), in +log space, already normalized / just used as-is
    :param am_seq_len: (batch,), int32
    :param edges: (4,num_edges), edges of the graph (from,to,emission_idx,sequence_idx)
    :param weights: (num_edges,), weights of the edges
    :param start_end_states: (2, batch), (start,end) state idx in automaton.
        there is only one single automaton.
    :param mask_idx: vocab index used for masking (e.g. padding, or if not path was found)
    :return: (alignment, scores), alignment is (time, batch), scores is (batch,), in +log space.
        note: scores are not differentiable here.
        do gather+sum on the am_scores by the alignment to get it differentiable.
    """
    last_state_idx = start_end_states[1].max()
    n_states = last_state_idx + 1
    maker = OpMaker(OpDescription.from_gen_base(native_op.FastViterbiOp))
    op = maker.make_op()
    alignment, scores = op(am_scores, am_seq_len, edges, weights, start_end_states, n_states, mask_idx)
    return alignment, scores


def ctc_best_path(
    *,
    logits: torch.Tensor,
    logits_seq_lens: torch.Tensor,
    targets: torch.Tensor,
    targets_seq_lens: torch.Tensor,
    label_loop: bool = True,
    logits_time_major: bool = False,
    logits_normalize: bool = True,
    blank_index: int = -1,
) -> torch.Tensor:
    """
    :param logits: (time,batch,dim) or (batch,time,dim). unnormalized (before softmax)
    :param logits_seq_lens: shape (batch,) of int32|int64
    :param logits_time_major:
    :param targets: batch-major, [batch,time]
    :param targets_seq_lens: (batch,)
    :param label_loop: (ctc_merge_repeated in tf.nn.ctc_loss)
    :param logits_normalize: apply log_softmax on logits (default).
      if False, you might also set grad_wrt_softmax_in=False
    :param blank_index: vocab index of the blank symbol
    :return: alignment, (time, batch). note, to get the scores, do gather+sum on the am_scores by the alignment.
    """
    assert logits.ndim == 3
    dim = logits.shape[-1]
    if not logits_time_major:
        logits = torch.transpose(logits, 0, 1)  # (time,batch,dim)
    if logits_normalize:
        log_sm = torch.log_softmax(logits, dim=-1)  # (time,batch,dim)
    else:
        log_sm = logits

    if blank_index < 0:
        blank_index += dim
    assert 0 <= blank_index < dim
    edges, weights, start_end_states = get_ctc_fsa_fast_bw(
        targets=targets, seq_lens=targets_seq_lens, blank_idx=blank_index, label_loop=label_loop
    )

    alignment, _ = fast_viterbi(
        am_scores=log_sm,
        am_seq_len=logits_seq_lens,
        edges=edges,
        weights=weights,
        start_end_states=start_end_states,
        mask_idx=blank_index,
    )
    return alignment


def edit_distance(a, a_len, b, b_len):
    """
    Wraps :class:`NativeOp.EditDistanceOp`.

    :param torch.Tensor a: (batch,time1), int32
    :param torch.Tensor a_len: (batch,), int32
    :param torch.Tensor b: (batch,time2), int32
    :param torch.Tensor b_len: (batch,), int32
    :return: (batch,) tensor, int32, un-normalized edit distance
    :rtype: torch.Tensor
    """
    maker = OpMaker(OpDescription.from_gen_base(native_op.EditDistanceOp))
    op = maker.make_op()
    return op(a, a_len, b, b_len)


def optimal_completion_edit_distance(a, a_len, b, b_len):
    """
    Wraps :class:`NativeOp.OptimalCompletionEditDistanceOp`.

    :param torch.Tensor a: (batch,time1), int32. prefix
    :param torch.Tensor a_len: (batch,), int32
    :param torch.Tensor b: (batch,time2), int32
    :param torch.Tensor b_len: (batch,), int32
    :return: (batch,) tensor, int32, un-normalized edit distance
    :rtype: torch.Tensor
    """
    maker = OpMaker(OpDescription.from_gen_base(native_op.OptimalCompletionEditDistanceOp))
    op = maker.make_op()
    return op(a, a_len, b, b_len)


def optimal_completion_edit_distance_per_successor(a, a_len, b, b_len, successors):
    """
    Wraps :class:`NativeOp.OptimalCompletionEditDistancePerSuccessorOp`.

    :param torch.Tensor a: (batch,time1), int32. prefix
    :param torch.Tensor a_len: (batch,), int32
    :param torch.Tensor b: (batch,time2), int32
    :param torch.Tensor b_len: (batch,), int32
    :param torch.Tensor|int successors: (n_labels,), int32. scalar means tf.range(successors)
    :return: (batch,n_labels) tensor, int32, un-normalized edit distance
    :rtype: torch.Tensor
    """
    if isinstance(successors, int):
        n_labels = successors
        successors = torch.arange(0, n_labels, 1)
    assert isinstance(successors, torch.Tensor)
    maker = OpMaker(OpDescription.from_gen_base(native_op.OptimalCompletionEditDistancePerSuccessorOp))
    op = maker.make_op()
    return op(a, a_len, b, b_len, successors)


def next_edit_distance_row(last_row, a, a_n, a_ended, b, b_len):
    """
    Wraps :class:`NativeOp.NextEditDistanceRowOp`.

    :param torch.Tensor last_row: 2d (batch,b_time + 1), int32. last edit distances
    :param torch.Tensor a: symbols. 1d (batch,), int32. current.
    :param torch.Tensor|int a_n: scalar or 1d (batch,), int32. current position
    :param torch.Tensor a_ended: 1d (batch,), int32 (casted from bool, because int32 easier to handle)
    :param torch.Tensor b: symbols. 2d (batch,b_time), int32
    :param torch.Tensor b_len: 1d (batch,), int32
    :return: 2d (batch,b_time + 1), int32, next (unnormalized) edit distance row
    :rtype: torch.Tensor
    """
    a_ended = a_ended.int()
    if isinstance(a_n, int):
        a_n = torch.tensor(a_n, device=a.device)
    if a_n.ndim == 0:
        a_n = a_n[None].tile(a_ended.shape)
    maker = OpMaker(OpDescription.from_gen_base(native_op.NextEditDistanceRowOp))
    op = maker.make_op()
    return op(last_row, a, a_n, a_ended, b, b_len)


def edit_distance_via_next_edit_distance_row(a, a_len, b, b_len, optimal_completion=False, full_row_output=False):
    """
    This is mostly for demonstration and debugging.
    Should be equivalent to :func:`edit_distance` or :func:`optimal_completion_edit_distance`
    (which should be much faster).

    :param torch.Tensor a: (batch,time1), int32
    :param torch.Tensor a_len: (batch,), int32
    :param torch.Tensor b: (batch,time2), int32
    :param torch.Tensor b_len: (batch,), int32
    :param bool optimal_completion: calc optimal completion edit distance instead
    :param bool full_row_output: outputs the full final row
    :return: (batch,) or (batch,time2+1) tensor, int32, un-normalized edit distance
    :rtype: torch.Tensor
    """
    batch_size = a.size(0)
    time1 = a.size(1)
    time2 = b.size(1)

    row = torch.arange(time2 + 1, device=a.device)[None, :].tile((batch_size, 1))  # (B,time2+1)
    for i in range(time1):
        a_ended = i >= a_len  # (B,)
        a_cur = a[:, i]  # (B,)
        row = next_edit_distance_row(a=a_cur, a_n=i, a_ended=a_ended, b=b, b_len=b_len, last_row=row)

    if full_row_output:
        assert not optimal_completion  # assert the default, this would not have an effect
        return row
    elif not optimal_completion:
        return row[:, -1]
    else:
        return torch.min(row, dim=1).values


def next_edit_distance_reduce(last_row, a, a_n, a_ended, b, b_len, optimal_completion=False, a_blank_idx=None):
    """
    Wraps :class:`NativeOp.NextEditDistanceReduceOp`.

    :param torch.Tensor last_row: 2d (batch,b_time + 1), int32. last edit distances
    :param torch.Tensor a: symbols. 2d (batch|1,n_labels), int32. current.
    :param torch.Tensor a_n: scalar or 1d (batch,), int32. current position
    :param torch.Tensor a_ended: 1d (batch,), int32 (casted from bool, because int32 easier to handle)
    :param torch.Tensor b: symbols. 2d (batch,b_time), int32
    :param torch.Tensor b_len: 1d (batch,), int32
    :param torch.Tensor|int|None a_blank_idx: scalar, int32
    :param bool|torch.Tensor optimal_completion:
    :return: 2d (batch,n_labels), int32, next (unnormalized) (optimal completion) edit distance
    :rtype: torch.Tensor
    """
    a_ended = a_ended.int()
    if a_n.ndim == 0:
        a_n = a_n[None].tile(a_ended.shape)
    if a_blank_idx is None:
        a_blank_idx = -1
    maker = OpMaker(OpDescription.from_gen_base(native_op.NextEditDistanceReduceOp))
    op = maker.make_op()
    return op(last_row, a, a_n, a_ended, b, b_len, optimal_completion, a_blank_idx)


def optimal_completion_edit_distance_per_successor_via_next_edit_distance(a, a_len, b, b_len, successors):
    """
    Uses :func:`next_edit_distance_reduce` and :func:`edit_distance_via_next_edit_distance_row`.
    Mostly for demonstration/testing.
    In practice, you would do something similar, but in your own loop.
    Similar to :func:`optimal_completion_edit_distance_per_successor`,
    but the handling of ended sequences (from ``a``) is different.

    :param torch.Tensor a: (batch,time1), int32. prefix
    :param torch.Tensor a_len: (batch,), int32
    :param torch.Tensor b: (batch,time2), int32
    :param torch.Tensor b_len: (batch,), int32
    :param torch.Tensor|int successors: (n_labels,), int32. scalar means tf.range(successors)
    :return: (batch,n_labels) tensor, int32, un-normalized edit distance
    :rtype: torch.Tensor
    """
    if isinstance(successors, int):
        n_labels = successors
        successors = torch.arange(0, n_labels, 1)[None]  # (1,n_labels)
    assert isinstance(successors, torch.Tensor)
    last_row = edit_distance_via_next_edit_distance_row(a, a_len, b, b_len, full_row_output=True)
    return next_edit_distance_reduce(
        last_row,
        a=successors,
        a_n=torch.tensor(a.size(1)),
        a_ended=a_len != a.size(1),
        b=b,
        b_len=b_len,
        optimal_completion=True,
    )



"""
Assertion utility functions for validating conditions in Python code.
"""

from __future__ import annotations
from typing import Union
import returnn.frontend as rf
from returnn.tensor import Tensor


__all__ = ["assert_"]


def assert_(condition: Union[Tensor, bool], message: str, *, stop: bool = True) -> None:
    """
    Asserts that a given condition is True.
    If the condition is False, raises an AssertionError with the provided message.
    This runs async on GPU.

    :param condition:
    :param message:
    :param stop: if True, raises an AssertionError on failure; if False, only logs a warning
    :return: nothing
    """
    if isinstance(condition, bool):
        if not condition:
            if stop:
                raise AssertionError(message)
            else:
                print(f"[ASSERT FAILED WARNING]: {message}")

    elif isinstance(condition, Tensor):
        if condition.dims:
            condition = rf.reduce_all(condition, axis=condition.dims)  # reduce to scalar
        # noinspection PyProtectedMember
        condition._raw_backend.assert_(condition, message=message, stop=stop)

    else:
        raise TypeError(f"Condition must be a boolean or a Tensor, got {type(condition)}")



"""
Helpers to improve torch.compile on RF code.
"""

from __future__ import annotations
from typing import Any, Iterable, List, Tuple

import os
from returnn.tensor import Tensor, Dim

# noinspection PyProtectedMember
from returnn.frontend import _native

_is_set_up = False


def setup():
    """
    Set up the torch.compile helpers for RF code, also including :class:`Tensor` and :class:`Dim`.
    """

    global _is_set_up
    if _is_set_up:
        return
    _is_set_up = True  # only try once

    assert not _native.is_set_up(), "Call this setup() as early as possible."
    _native.set_enabled(False)

    # We have lots of dynamic shapes.
    os.environ["TORCHDYNAMO_CAPTURE_SCALAR_OUTPUTS"] = "1"

    # noinspection PyProtectedMember
    from torch.utils._pytree import register_pytree_node

    register_pytree_node(Tensor, _tensor_flatten, _tensor_unflatten)
    register_pytree_node(Dim, _dim_flatten, _dim_unflatten)

    Dim.get_dim_value = _dim_get_dim_value


def _tensor_flatten(t: Tensor) -> Tuple[List[Any], Any]:
    """
    Flatten the tensor for PyTree.
    """
    return [t.raw_tensor, t.dims, t.sparse_dim], [
        t.name,
        t.dtype,
        t.version,
        t.feature_dim_axis_or_unspecified,
        t.time_dim_axis_or_unspecified,
    ]


def _tensor_unflatten(values: Iterable[Any], metadata: Any) -> Tensor:
    """
    Unflatten the tensor from PyTree.
    """
    raw_tensor, dims, sparse_dim = values
    name, dtype, version, feature_dim_axis, time_dim_axis = metadata
    return Tensor(
        name=name,
        dims=dims,
        dtype=dtype,
        sparse_dim=sparse_dim,
        feature_dim_axis=feature_dim_axis,
        time_dim_axis=time_dim_axis,
        raw_tensor=raw_tensor,
        version=version,
    )


def _dim_flatten(d: Dim) -> Tuple[List[Any], Any]:
    """
    Flatten the dim for PyTree.
    """
    return [d.dyn_size_ext], [d.name, d.dimension, d.size]


def _dim_unflatten(values: Iterable[Any], metadata: Any) -> Dim:
    """
    Unflatten the dim from PyTree.
    """
    (dyn_size_ext,) = values
    name, dimension, size = metadata
    # TODO this creates a new instance... this is maybe wrong?
    return Dim(name=name, dimension=dimension, size=size, dyn_size_ext=dyn_size_ext)


def _dim_get_dim_value(self: Dim) -> int:
    """
    Infers the dim this axis should have if unbroadcasted.
    If `self.src_data` has a placeholder, will use the shape from there.
    Otherwise, uses `self.dimension` (if static) or `self.dyn_size` (if dynamic).

    :return: max(size or dyn_size)
    """
    res = self.get_dim_value_tensor()
    if isinstance(res, Tensor):
        assert res.dims == ()
        assert res.raw_tensor is not None
        # Specifically PyTorch would then treat it as a SymInt in torch.compile,
        # which is important to have for some torch functions (e.g. torch.tile and others).
        return int(res.raw_tensor)
    assert isinstance(res, int)
    return res



"""
Array (Tensor) functions
"""

from __future__ import annotations
from typing import Optional, Union
import torch


# noinspection PyShadowingBuiltins
def masked_select(input: torch.Tensor, mask: torch.Tensor, *, mask_len: Optional[Union[int, torch.Tensor]] = None):
    """
    Like :func:`torch.masked_select` but much more efficient,
    both in terms of memory and computation time,
    both on CPU and GPU.

    See here for the issues with :func:`torch.masked_select`:
    https://github.com/rwth-i6/returnn/issues/1584
    https://github.com/pytorch/pytorch/issues/30246
    https://github.com/pytorch/pytorch/issues/56896

    :param input: [mask_dims..., remaining_dims...]
    :param mask: [mask_dims...], binary mask to index with. if it has less dims than ``input``,
        the remaining dims are broadcasted.
    :param mask_len: if given, the length of the mask. this avoids a CUDA synchronization.
    :return: selected elements, shape [mask_len, remaining_dims...]
    """
    assert input.ndim >= mask.ndim
    assert all(input.shape[i] == mask.shape[i] for i in range(mask.ndim))
    mask_flat = mask.flatten()
    # Note: So far it seems that our custom nonzero is always slower than torch.nonzero,
    # thus we always use torch.nonzero here for now.
    # https://github.com/rwth-i6/returnn/pull/1593
    # We might change this in the future. See also:
    # https://github.com/pytorch/pytorch/issues/131256
    indices = torch.nonzero(mask_flat).squeeze(1)  # [out_len]
    if mask_len is not None:
        assert indices.shape[0] == mask_len
    input_flat = input.flatten(end_dim=mask.ndim - 1)
    return input_flat[indices]


def nonzero(mask: torch.Tensor, *, out_len: Union[int, torch.Tensor]) -> torch.Tensor:
    """
    This has the advantage over :func:`torch.nonzero`
    that we do not need to perform a CUDA synchronization.
    We can avoid that when we know the output length in advance.

    However, in my benchmarks, it seems this is slower than torch.nonzero.
    https://github.com/rwth-i6/returnn/pull/1593
    https://github.com/pytorch/pytorch/issues/131256

    :param mask: flattened (dim() == 1) mask, bool
    :param out_len:
    :return: indices of True elements, shape [out_len].
        like ``mask.nonzero().flatten()``
    """
    assert mask.dim() == 1 and mask.dtype == torch.bool
    # Sort currently does not support bool dtype on CUDA, thus cast to int.
    idx = torch.argsort(mask.to(torch.int8), stable=True, descending=True)  # [in_len]
    idx = idx[:out_len]  # [out_len]
    return idx


def sequence_mask(lengths: torch.Tensor, *, maxlen: Optional[int] = None) -> torch.Tensor:
    """
    Creates a boolean mask from sequence lengths.

    :param lengths: Tensor of shape [batch_size...] containing sequence lengths
    :param maxlen: Maximum length of the sequences. If None, uses the maximum value in lengths.
    :return: A boolean mask tensor of shape [batch_size..., maxlen]
    """
    if maxlen is None:
        maxlen = lengths.max()
    indices = torch.arange(0, maxlen, dtype=lengths.dtype, device=lengths.device)
    mask = indices < lengths[..., None]
    return mask


def sequence_mask_time_major(lengths: torch.Tensor, *, maxlen: Optional[int] = None) -> torch.Tensor:
    """
    Creates a boolean mask from sequence lengths.

    :param lengths: Tensor of shape [batch_size...] containing sequence lengths
    :param maxlen: Maximum length of the sequences. If None, uses the maximum value in lengths.
    :return: A boolean mask tensor of shape [maxlen, batch_size...]
    """
    if maxlen is None:
        maxlen = lengths.max()
    indices = torch.arange(0, maxlen, dtype=lengths.dtype, device=lengths.device)
    mask = indices[(slice(None),) + (None,) * lengths.ndim] < lengths[None]
    return mask



"""
Allow to use Numpy arrays.
This backend will probably never be feature-complete.
It just has the bare minimum such that the user can assign Numpy arrays to Tensor.raw_tensor.
"""

from __future__ import annotations
from typing import Optional, Union, Sequence, Tuple
import numpy
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import Backend
from returnn.frontend import RawTensorTypes


# We do not expect that we will ever implement all the methods of the Backend interface.
# noinspection PyAbstractClass
class NumpyBackend(Backend[numpy.ndarray]):
    """Numpy backend"""

    name = "numpy"
    RawTensorType = numpy.ndarray

    @staticmethod
    def executing_eagerly() -> bool:
        """executing eagerly"""
        return True

    @staticmethod
    def assert_(condition: Tensor, message: str, *, stop: bool = True):
        """assert"""
        assert condition.dims == (), "condition for assert must be a scalar"
        if not condition.raw_tensor.item():
            if stop:
                raise AssertionError(message)
            else:
                print(f"[ASSERT FAILED WARNING]: {message}")

    @staticmethod
    def get_dtype_name_raw(raw_tensor: numpy.ndarray) -> str:
        """
        :return: dtype of raw tensor, as string. e.g. "int64" etc.
        """
        dtype_name = raw_tensor.dtype.name
        # See returnn.datasets.util.strings.str_to_numpy_array.
        if dtype_name.startswith("str"):
            return "string"  # ignore the bit-length, it supports variable length
        return dtype_name

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> numpy.dtype:
        """
        :param dtype_name: e.g. "float32"
        :return: dtype object
        """
        return numpy.dtype(dtype_name)

    @staticmethod
    def get_ndim_raw(raw_tensor: numpy.ndarray) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        return raw_tensor.ndim

    @staticmethod
    def get_shape_raw(raw_tensor: numpy.ndarray) -> Tuple[int, ...]:
        """
        :return: shape of raw tensor
        """
        return raw_tensor.shape

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: numpy.ndarray) -> Tuple[int, ...]:
        """
        :return: shape of raw tensor. assumes that ndim is known.
            In eager frameworks, all dims are int.
        """
        return raw_tensor.shape

    @staticmethod
    def get_known_shape_raw(raw_tensor: numpy.ndarray) -> Tuple[int, ...]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
            This will not create any ops.
            In eager frameworks, all dims are known.
        """
        return raw_tensor.shape

    @staticmethod
    def convert_to_tensor(
        value: Union[Tensor, numpy.ndarray, RawTensorTypes],
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
        device: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Tensor[numpy.ndarray]:
        """convert to tensor"""
        if isinstance(value, Tensor):
            return value
        if isinstance(value, numpy.ndarray):
            name = name or "raw_tensor"
        else:
            name = name or "const"
            value = numpy.array(value, dtype=NumpyBackend.as_dtype_raw(dtype))
        assert isinstance(value, numpy.ndarray)
        return Tensor(name, dims=dims, dtype=dtype, sparse_dim=sparse_dim, feature_dim=feature_dim, raw_tensor=value)

    @staticmethod
    def expand_dims_raw(raw_tensor: numpy.ndarray, axis: int) -> numpy.ndarray:
        """
        :param raw_tensor:
        :param axis: e.g. 1
        :return: raw tensor with new axis
        """
        return numpy.expand_dims(raw_tensor, axis)

    @staticmethod
    def transpose_raw(raw_tensor: numpy.ndarray, perm: Sequence[int]) -> numpy.ndarray:
        """
        :param raw_tensor:
        :param perm: e.g. [0, 2, 1]
        :return: permuted (transposed) raw tensor
        """
        if all(p == i for i, p in enumerate(perm)):
            return raw_tensor
        return raw_tensor.transpose(tuple(perm))

    @staticmethod
    def reshape_raw(
        raw_tensor: numpy.ndarray, shape: Union[Sequence[Union[int, numpy.ndarray]], numpy.ndarray]
    ) -> numpy.ndarray:
        """reshape raw"""
        return numpy.reshape(raw_tensor, shape)

    @staticmethod
    def compare_raw(a: numpy.ndarray, kind: str, b: numpy.ndarray) -> numpy.ndarray:
        """
        :param a:
        :param kind: "equal", "less", "less_equal", "greater", "greater_equal", "not_equal"
        :param b:
        :return: a `kind` b
        """
        assert a.ndim == b.ndim or a.ndim == 0 or b.ndim == 0
        op = getattr(numpy, kind)  # e.g. numpy.equal
        return op(a, b)

    _CombineKindMap = {"mul": numpy.multiply}

    @staticmethod
    def combine_raw(a: numpy.ndarray, kind: str, b: numpy.ndarray) -> numpy.ndarray:
        """
        :param a:
        :param kind: "add", "sub", "mul", "truediv", "floordiv", "mod", "pow",
            "maximum", "minimum", "logical_and", "logical_or", "squared_difference"
        :param b:
        :return: a `kind` b
        """
        assert a.ndim == b.ndim or a.ndim == 0 or b.ndim == 0
        op = getattr(numpy, kind, None)  # e.g. numpy.add
        if not op:
            op = NumpyBackend._CombineKindMap.get(kind)
            if not op:
                raise ValueError(f"RF NumpyBackend: combine kind {kind!r} not supported")
        res = op(a, b)
        if not isinstance(res, numpy.ndarray):
            res = numpy.array(res)
        return res

    @staticmethod
    def where(
        cond: Tensor,
        true_: Union[Tensor, rf.RawTensorTypes],
        false_: Union[Tensor, rf.RawTensorTypes],
        *,
        allow_broadcast_all_sources: bool = False,
    ) -> Tensor:
        """where"""
        if isinstance(true_, Tensor):
            dtype = true_.dtype
        elif isinstance(false_, Tensor):
            dtype = false_.dtype
        else:
            dtype = None
        true_ = rf.convert_to_tensor(true_, _backend=NumpyBackend, dtype=dtype)
        false_ = rf.convert_to_tensor(false_, _backend=NumpyBackend, dtype=dtype)
        out = Tensor.get_common_data(
            [true_, false_, cond], allow_broadcast_all_sources=allow_broadcast_all_sources, name="where"
        )
        out.dtype = true_.dtype
        out.sparse_dim = true_.sparse_dim or false_.sparse_dim
        out.feature_dim = true_.feature_dim or false_.feature_dim
        cond_bc_raw = cond.copy_compatible_to_dims_raw(out.dims)
        true_bc_raw = true_.copy_compatible_to_dims_raw(out.dims)
        false_bc_raw = false_.copy_compatible_to_dims_raw(out.dims)
        out.raw_tensor = numpy.where(cond_bc_raw, true_bc_raw, false_bc_raw)
        return out

    @staticmethod
    def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor[numpy.ndarray]:
        """
        :param dim:
        :param dtype:
        :param device:
        :return: tensor with shape [dim]
        """
        if not dtype and dim.dyn_size_ext is not None:
            dtype = dim.dyn_size_ext.dtype
        if not dtype:
            dtype = rf.get_default_array_index_dtype()
        out = Tensor(
            "range",
            dims=[dim],
            sparse_dim=dim if dtype.startswith("int") or dtype.startswith("uint") else None,
            dtype=dtype,
        )
        out.raw_tensor = numpy.arange(dim.get_dim_value(), dtype=NumpyBackend.as_dtype_raw(out.dtype))
        return out

    @staticmethod
    def reduce(
        source: Tensor[numpy.ndarray],
        *,
        mode: str,
        axis: Union[Dim, Sequence[Dim]],
        use_mask: bool = True,
    ) -> Tensor[numpy.ndarray]:
        """reduce"""
        assert mode in Backend._AllowedReduceModes
        if use_mask:
            if isinstance(axis, Dim):
                assert not axis.need_masking()  # not implemented
            else:
                assert all(not dim.need_masking() for dim in axis)  # not implemented
        func = getattr(numpy, mode)
        raw_dims = (
            [source.get_axis_from_description(axis)]
            if isinstance(axis, Dim)
            else [source.get_axis_from_description(dim) for dim in axis]
        )
        res_dims = [dim for i, dim in enumerate(source.dims) if i not in raw_dims]
        if not res_dims:
            # All are reduced. Need numpy.array() to get a tensor again.
            raw_result = numpy.array(func(source.raw_tensor))
        else:
            raw_result = func(source.raw_tensor, axis=raw_dims)
        res = Tensor(
            name=f"reduce_{mode}",
            raw_tensor=raw_result,
            dims=res_dims,
            dtype=source.dtype,
            sparse_dim=source.sparse_dim,
        )
        return res

    @staticmethod
    def activation_raw(raw_tensor: numpy.ndarray, func: str) -> numpy.ndarray:
        """
        :param raw_tensor:
        :param func: "tanh", "sigmoid", "relu", ...
        :return: raw tensor with elementwise activation applied
        """
        if func == "relu":
            return numpy.array(numpy.maximum(raw_tensor, 0))
        raise NotImplementedError("NumpyBackend: activation %r not implemented" % func)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim, single_step_dim
from rf_utils import run_model


def test_lstm():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "state_h": Tensor("state_h", [batch_dim, out_dim], dtype="float32"),
            "state_c": Tensor("state_c", [batch_dim, out_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.lstm = rf.LSTM(in_dim, out_dim)
            # better for the test to also have some random values in the bias, not zeros
            self.lstm.bias.initial = rf.init.Glorot()

        def __call__(self, x: Tensor, *, spatial_dim: Dim, state: rf.LstmState) -> Tuple[Tensor, rf.LstmState]:
            return self.lstm(x, state=state, spatial_dim=spatial_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        state = rf.LstmState(h=extern_data["state_h"], c=extern_data["state_c"])
        out, new_state = model(extern_data["data"], state=state, spatial_dim=time_dim)
        out.mark_as_output("out", shape=(batch_dim, time_dim, out_dim))
        new_state.h.mark_as_output("h", shape=(batch_dim, out_dim))
        new_state.c.mark_as_output("c", shape=(batch_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_lstm_single_step():
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, in_dim], dtype="float32"),
            "state_h": Tensor("state_h", [batch_dim, out_dim], dtype="float32"),
            "state_c": Tensor("state_c", [batch_dim, out_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.lstm = rf.LSTM(in_dim, out_dim)
            # better for the test to also have some random values in the bias, not zeros
            self.lstm.bias.initial = rf.init.Glorot()

        def __call__(self, x: Tensor, *, spatial_dim: Dim, state: rf.LstmState) -> Tuple[Tensor, rf.LstmState]:
            return self.lstm(x, state=state, spatial_dim=single_step_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        state = rf.LstmState(h=extern_data["state_h"], c=extern_data["state_c"])
        out, new_state = model(extern_data["data"], state=state, spatial_dim=single_step_dim)
        out.mark_as_output("out", shape=(batch_dim, out_dim))
        new_state.h.mark_as_output("h", shape=(batch_dim, out_dim))
        new_state.c.mark_as_output("c", shape=(batch_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_zoneout_lstm():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "state_h": Tensor("state_h", [batch_dim, out_dim], dtype="float32", feature_dim=out_dim),
            "state_c": Tensor("state_c", [batch_dim, out_dim], dtype="float32", feature_dim=out_dim),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.lstm = rf.ZoneoutLSTM(
                in_dim,
                out_dim,
                zoneout_factor_cell=0.15,
                zoneout_factor_output=0.05,
            )
            # better for the test to also have some random values in the bias, not zeros
            self.lstm.bias.initial = rf.init.Glorot()

        def __call__(self, x: Tensor, *, spatial_dim: Dim, state: rf.LstmState) -> Tuple[Tensor, rf.LstmState]:
            return self.lstm(x, state=state, spatial_dim=spatial_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        state = rf.LstmState(h=extern_data["state_h"], c=extern_data["state_c"])
        out, new_state = model(extern_data["data"], state=state, spatial_dim=time_dim)
        out.mark_as_output("out", shape=(batch_dim, time_dim, out_dim))
        new_state.h.mark_as_output("h", shape=(batch_dim, out_dim))
        new_state.c.mark_as_output("c", shape=(batch_dim, out_dim))

    # TODO ... TF needs TensorArray support in the internal rf.scan, not yet implemented
    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step, test_tensorflow=False)


def test_zoneout_lstm_single_step():
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, in_dim], dtype="float32"),
            "state_h": Tensor("state_h", [batch_dim, out_dim], dtype="float32"),
            "state_c": Tensor("state_c", [batch_dim, out_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.lstm = rf.ZoneoutLSTM(
                in_dim,
                out_dim,
                zoneout_factor_cell=0.15,
                zoneout_factor_output=0.05,
            )
            # better for the test to also have some random values in the bias, not zeros
            self.lstm.bias.initial = rf.init.Glorot()

        def __call__(self, x: Tensor, *, spatial_dim: Dim, state: rf.LstmState) -> Tuple[Tensor, rf.LstmState]:
            return self.lstm(x, state=state, spatial_dim=single_step_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        state = rf.LstmState(h=extern_data["state_h"], c=extern_data["state_c"])
        out, new_state = model(extern_data["data"], state=state, spatial_dim=single_step_dim)
        out.mark_as_output("out", shape=(batch_dim, out_dim))
        new_state.h.mark_as_output("h", shape=(batch_dim, out_dim))
        new_state.c.mark_as_output("c", shape=(batch_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_zoneout_lstm_tf_layers_vs_rf_pt():
    # Compare TF-layers ZoneoutLSTM vs RF ZoneoutLSTM with PyTorch backend.
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    # Note: Use high dimension because only that triggers some inconsistencies.
    in_dim, out_dim = Dim(7, name="in"), Dim(100, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    zoneout_factor_cell = 0.15
    zoneout_factor_output = 0.05

    tf_net_dict = {
        "lstm": {
            "class": "rec",
            "unit": "ZoneoutLSTM",
            "from": "data",
            "unit_opts": {
                "zoneout_factor_cell": zoneout_factor_cell,
                "zoneout_factor_output": zoneout_factor_output,
                "use_zoneout_output": True,
            },
            "in_dim": in_dim,
            "out_dim": out_dim,
            # Better for the test to not have zeros in weights/bias.
            "forward_weights_init": "random_normal_initializer(mean=0.0, stddev=1.)",
            "recurrent_weights_init": "random_normal_initializer(mean=0.0, stddev=1.)",
            "bias_init": "random_normal_initializer(mean=0.0, stddev=1.)",
        },
        "output": {"class": "copy", "from": "lstm"},
    }

    from returnn.config import Config

    config = Config(
        dict(
            log_verbositiy=5,
            network=tf_net_dict,
            extern_data={
                "data": {"dim_tags": extern_data["data"].dim_tags},
            },
        )
    )

    from returnn.tensor.utils import tensor_dict_fill_random_numpy_
    from returnn.torch.data.tensor_utils import tensor_dict_numpy_to_torch_
    from returnn.tf.network import TFNetwork
    import tensorflow as tf
    import numpy

    print("*** Construct TF graph for TF model")
    extern_data = TensorDict()
    extern_data.update(config.typed_dict["extern_data"], auto_convert=True)
    tensor_dict_fill_random_numpy_(extern_data, dyn_dim_max_sizes={time_dim: 20}, dyn_dim_min_sizes={time_dim: 10})
    extern_data_numpy_raw_dict = extern_data.as_raw_tensor_dict()
    extern_data.reset_content()
    rnd = numpy.random.RandomState(43)

    rf.select_backend_returnn_layers_tf()
    tf1 = tf.compat.v1
    with tf1.Graph().as_default() as graph, tf1.Session(graph=graph).as_default() as session:
        net = TFNetwork(config=config)
        net.construct_from_dict(config.typed_dict["network"])
        print("*** Random init TF model")
        net.initialize_params(session)

        # Overwrite again, such that the bias is also non-zero.
        print("*** Set random weights")
        params = net.get_params_list()
        for param in params:
            print("Param:", param, "shape:", param.shape, "dtype:", param.dtype)
            param.load(rnd.normal(scale=0.1, size=tuple(param.shape)), session=session)

        print("*** Forward")
        extern_data_tf_raw_dict = net.extern_data.as_raw_tensor_dict()
        assert set(extern_data_tf_raw_dict.keys()) == set(extern_data_numpy_raw_dict.keys())
        feed_dict = {extern_data_tf_raw_dict[k]: extern_data_numpy_raw_dict[k] for k in extern_data_numpy_raw_dict}
        fetches = net.get_fetches_dict()
        old_model_outputs_data = {}
        for old_layer_name in ["output"]:
            layer = net.get_layer(old_layer_name)
            out = layer.output
            old_model_outputs_data[old_layer_name] = out
            fetches["layer:" + old_layer_name] = out.placeholder
            for i, tag in enumerate(out.dim_tags):
                if tag.is_batch_dim():
                    fetches[f"layer:{old_layer_name}:size{i}"] = tag.get_dim_value()
                elif tag.dyn_size_ext is not None:
                    old_model_outputs_data[f"{old_layer_name}:size{i}"] = tag.dyn_size_ext
                    fetches[f"layer:{old_layer_name}:size{i}"] = tag.dyn_size_ext.placeholder
        old_model_outputs_fetch = session.run(fetches, feed_dict=feed_dict)
        old_model_params = net.get_params_serialized(session=session)

    # --- Now the PyTorch model
    import torch

    print("*** Create PyTorch model")
    rf.select_backend_torch()

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.lstm = rf.ZoneoutLSTM(
                in_dim,
                out_dim,
                zoneout_factor_cell=zoneout_factor_cell,
                zoneout_factor_output=zoneout_factor_output,
                forget_bias=1.0,  # RETURNN TF-layers ZoneoutLSTM
                parts_order="icfo",  # RETURNN TF-layers ZoneoutLSTM
            )

        def __call__(self, x: Tensor, *, spatial_dim: Dim) -> Tensor:
            initial_state = self.lstm.default_initial_state(batch_dims=[batch_dim])
            y, _ = self.lstm(x, state=initial_state, spatial_dim=spatial_dim)
            return y

    new_model = _Net()

    print("*** Copy over params")
    tf_lstm_kernel = old_model_params.values_dict["lstm"]["lstm_cell/kernel"]  # (in+out, 4*out)
    tf_lstm_bias = old_model_params.values_dict["lstm"]["lstm_cell/bias"]  # (4*out,)
    tf_lstm_kernel = tf_lstm_kernel.transpose([1, 0])  # (4*out, in+out)
    tf_lstm_ff_kernel = tf_lstm_kernel[:, : in_dim.dimension]
    tf_lstm_rec_kernel = tf_lstm_kernel[:, in_dim.dimension :]
    with torch.no_grad():
        new_model.lstm.ff_weight.raw_tensor.copy_(torch.from_numpy(tf_lstm_ff_kernel))
        new_model.lstm.rec_weight.raw_tensor.copy_(torch.from_numpy(tf_lstm_rec_kernel))
        new_model.lstm.bias.raw_tensor.copy_(torch.from_numpy(tf_lstm_bias))

    print("*** Forward")
    rf.init_train_step_run_ctx(train_flag=False, step=0, epoch=1)
    extern_data.reset_content()
    extern_data.assign_from_raw_tensor_dict_(extern_data_numpy_raw_dict)
    tensor_dict_numpy_to_torch_(extern_data)

    out = new_model(extern_data["data"], spatial_dim=time_dim)
    assert out.dims == (time_dim, batch_dim, out_dim) == old_model_outputs_data["output"].dims
    assert out.dtype == "float32" == old_model_outputs_data["output"].dtype

    print("*** Compare")
    import numpy.testing

    tf_out = old_model_outputs_fetch["layer:output"]
    pt_out = out.raw_tensor.detach().cpu().numpy()
    print("out shape:", tf_out.shape)
    assert tf_out.shape == pt_out.shape

    size_v = old_model_outputs_fetch[f"layer:output:size0"]
    print("* sizes:", size_v)
    for b in range(tf_out.shape[1]):
        tf_out[size_v[b] :] = 0
        pt_out[size_v[b] :] = 0

    _pt_out = pt_out
    _tf_out = tf_out

    for t in range(_tf_out.shape[0]):
        print("* t:", t)

        tf_out = _tf_out[t]
        pt_out = _pt_out[t]

        # Using equal_nan=False because we do not want any nan in any of the values.
        if numpy.allclose(tf_out, pt_out, atol=1e-5):
            continue  # done

        print("** not all close!")
        print("close:")
        # Iterate over all indices, and check if the values are close.
        # If not, add the index to the mismatches list.
        remarks = []
        count_mismatches = 0
        for idx in sorted(numpy.ndindex(tf_out.shape), key=sum):
            if numpy.isnan(tf_out[idx]) and numpy.isnan(pt_out[idx]):
                remarks.append("[%s]:? (both are nan)" % ",".join([str(i) for i in idx]))
                count_mismatches += 1
                continue
            close = numpy.allclose(tf_out[idx], pt_out[idx], atol=1e-5)
            if not close:
                count_mismatches += 1
            remarks.append(
                "[%s]:" % ",".join([str(i) for i in idx])
                + ("✓" if close else "✗ (%.5f diff)" % abs(tf_out[idx] - pt_out[idx]))
            )
            if len(remarks) >= 50 and count_mismatches > 0:
                remarks.append("...")
                break
        print("\n".join(remarks))
        numpy.testing.assert_allclose(
            tf_out,
            pt_out,
            rtol=1e-5,
            atol=1e-5,
            equal_nan=False,
            err_msg=f"TF-layers vs RF-PT mismatch",
        )
        raise Exception(f"should not get here, mismatches: {remarks}")

    print("all close!")


if __name__ == "__main__":
    import sys
    import unittest
    from returnn.util import better_exchook

    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
